import pandas as pd
import numpy as np
import pickle
import re

from icet import ClusterSpace
from icet.core.structure import Structure
from typing import List, Union
from ase import Atoms


class ClusterExpansion:
    """
    Cluster expansion model

    Attributes
    ----------
    cluster_space : icet.ClusterSpace
        cluster space that was used for constructing the cluster expansion
    parameters : numpy.ndarray
        effective cluster interactions (ECIs)
    """

    def __init__(self, cluster_space: ClusterSpace,
                 parameters: np.array) -> None:
        """
        Initializes a ClusterExpansion object.

        Parameters
        ----------
        cluster_space
            cluster space to be used for constructing the cluster expansion
        parameters
            effective cluster interactions (ECIs)

        Raises
        ------
        ValueError
            if cluster space and parameters differ in length
        """
        if len(cluster_space) != len(parameters):
            raise ValueError('cluster_space ({}) and parameters ({}) must have'
                             ' the same length'.format(len(cluster_space),
                                                       len(parameters)))
        self._cluster_space = cluster_space
        if isinstance(parameters, list):
            parameters = np.array(parameters)
        self._parameters = parameters
        self._original_parameters = parameters.copy()
        self._pruning_history = []

    def predict(self, structure: Union[Atoms, Structure]) -> float:
        """
        Predicts the property of interest (e.g., the energy) for the input
        structure using the cluster expansion.

        Parameters
        ----------
        structure
            atomic configuration

        Returns
        -------
        float
            property value of predicted by the cluster expansion
        """
        cluster_vector = self.cluster_space.get_cluster_vector(structure)
        prop = np.dot(cluster_vector, self.parameters)
        return prop

    @property
    def parameters_as_dataframe(self) -> pd.DataFrame:
        """ dataframe containing orbit data and ECIs """
        rows = self.cluster_space.orbit_data
        for row, eci in zip(rows, self.parameters):
            row['eci'] = eci
        return pd.DataFrame(rows)

    @property
    def orders(self) -> List:
        return list(range(len(self.cluster_space.cutoffs)+2))

    @property
    def cluster_space(self) -> ClusterSpace:
        """ cluster space on which cluster expansion is based """
        return self._cluster_space

    @property
    def parameters(self) -> List[float]:
        """ effective cluster interactions (ECIs) """
        return self._parameters

    def __len__(self) -> int:
        return len(self._parameters)

    def _get_string_representation(self, print_threshold: int = None,
                                   print_minimum: int = 10):
        """ String representation of the cluster expansion. """
        cluster_space_repr = self._cluster_space._get_string_representation(
            print_threshold, print_minimum).split('\n')
        # rescale width
        eci_col_width = max(
            len('{:9.3g}'.format(max(self._parameters, key=abs))), len('ECI'))
        width = len(cluster_space_repr[0]) + len(' | ') + eci_col_width

        s = []  # type: List
        s += ['{s:=^{n}}'.format(s=' Cluster Expansion ', n=width)]
        s += [t for t in cluster_space_repr if re.search(':', t)]

        # table header
        s += [''.center(width, '-')]
        t = [t for t in cluster_space_repr if 'index' in t]
        t += ['{s:^{n}}'.format(s='ECI', n=eci_col_width)]
        s += [' | '.join(t)]
        s += [''.center(width, '-')]

        # table body
        index = 0
        while index < len(self):
            if (print_threshold is not None and
                    len(self) > print_threshold and
                    index >= print_minimum and
                    index <= len(self) - print_minimum):
                index = len(self) - print_minimum
                s += [' ...']
            pattern = r'^{:4}'.format(index)
            t = [t for t in cluster_space_repr if re.match(pattern, t)]
            eci_value = '{:9.3g}'.format(self._parameters[index])
            t += ['{s:^{n}}'.format(s=eci_value, n=eci_col_width)]
            s += [' | '.join(t)]
            index += 1
        s += [''.center(width, '=')]

        return '\n'.join(s)

    def __repr__(self) -> str:
        """ string representation """
        return self._get_string_representation(print_threshold=50)

    def prune(self, indices: List[int] = None, tol: float = 0):
        """
        Removes orbits from the cluster expansion (CE), for which the effective
        cluster interactions (ECIs; parameters) are zero or close to zero.
        This commonly reduces the computational cost for evaluating the CE and
        is therefore recommended prior to using it in production. If the method
        is called without arguments orbits will be pruned, for which the ECIs
        are strictly zero. Less restrictive pruning can be achived by setting
        the `tol` keyword.

        Parameters
        ----------
        indices
            indices to parameters to remove in the cluster expansion.
        tol
            orbits for which the absolute ECIs is/are within this
            value will be pruned
        """
        self._pruning_history.append({'indices': indices, 'tol': tol})

        if indices is None:
            indices = [i for i, param in enumerate(
                self.parameters) if np.abs(param) <= tol and i > 0]
        df = self.parameters_as_dataframe
        indices = list(set(indices))

        if 0 in indices:
            raise ValueError('Orbit index cannot be 0 since'
                             ' the zerolet may not be pruned.')
        orbit_candidates_for_removal = \
            df.orbit_index[np.array(indices)].tolist()
        safe_to_remove_orbits, safe_to_remove_params = [], []
        for oi in set(orbit_candidates_for_removal):
            if oi == -1:
                continue
            orbit_count = df.orbit_index.tolist().count(oi)
            oi_remove_count = orbit_candidates_for_removal.count(oi)
            if orbit_count <= oi_remove_count:
                safe_to_remove_orbits.append(oi)
                safe_to_remove_params += df.index[df['orbit_index']
                                                  == oi].tolist()

        self._cluster_space._prune_orbit_list(indices=safe_to_remove_orbits)
        self._parameters = self._parameters[np.setdiff1d(
            np.arange(len(self._parameters)), safe_to_remove_params)]

    def write(self, filename: str):
        """
        Writes ClusterExpansion object to file.

        Parameters
        ---------
        filename
            name of file to which to write
        """
        self.cluster_space.write(filename)

        with open(filename, 'rb') as handle:
            data = pickle.load(handle)

        data['parameters'] = self.parameters
        data['original_parameters'] = self._original_parameters
        data['pruning_history'] = self._pruning_history

        with open(filename, 'wb') as handle:
            pickle.dump(data, handle, protocol=pickle.HIGHEST_PROTOCOL)

    @staticmethod
    def read(filename: str):
        """
        Reads ClusterExpansion object from file.

        Parameters
        ---------
        filename
            file from which to read
        """
        cs = ClusterSpace.read(filename)
        with open(filename, 'rb') as handle:
            data = pickle.load(handle)
        parameters = data['parameters']
        pruning_history = data['pruning_history']
        original_parameters = data['original_parameters']
        ce = ClusterExpansion(cs, original_parameters)
        for arg in pruning_history:
            ce.prune(indices=arg['indices'], tol=arg['tol'])

        assert list(parameters) == list(ce.parameters)
        return ce



from _icet import NeighborList
from .structure import Structure
from ase import Atoms


def get_neighbor_lists(atoms, cutoffs=None):
    '''
    Returns list of icet neighbor lists from a configuration and cutoffs.

    Parameters
    ----------
    atoms : ASE Atoms object / icet Structure object (bi-optional)
        atomic configuration
    cutoffs:
        positive floats indicating the cutoffs for the various clusters

    Returns
    -------
    list of NeighborList objects
    '''

    # deal with different types of structure objects
    if isinstance(atoms, Atoms):
        structure = Structure.from_atoms(atoms)
    elif isinstance(atoms, Structure):
        structure = atoms
    else:
        msg = ['Unknown structure format']
        msg += ['{} (ClusterSpace)'.format(type(atoms))]
        raise Exception(' '.join(msg))

    neighbor_lists = []
    if cutoffs is None:
        raise Exception('Both n and cutoffs is None in count clusters')
    else:
        for cutoff in cutoffs:
            nl = NeighborList(cutoff)
            neighbor_lists.append(nl)

    # build the neighbor_lists
    for nl in neighbor_lists:
        nl.build(structure)

    return neighbor_lists



"""
Definition of the variance-constrained semi-grand canonical ensemble class.
"""

import numpy as np

from ase import Atoms
from ase.data import atomic_numbers, chemical_symbols
from ase.units import kB
from typing import Dict, Union

from .. import DataContainer
from .base_ensemble import BaseEnsemble
from ..calculators.base_calculator import BaseCalculator


class VCSGCEnsemble(BaseEnsemble):
    """Instances of this class allow one to simulate systems in the
    variance-constrained semi-grand canonical (VCSGC) ensemble
    (:math:`N\\phi\\kappa VT`), i.e. at constant temperature (:math:`T`), total
    number of sites (:math:`N=\\sum_i N_i`), and two additional dimensionless
    parameters :math:`\\phi` and :math:`\\kappa`, which constrain average and
    variance of the concentration, respectively. The VCSGC ensemble is
    currently only implemented for binary systems.

    The probability for a particular state in the VCSGC ensemble for a
    :math:`2`-component system can be written

    .. math::

        \\rho_{\\text{VCSGC}} \\propto \\exp\\Big[ - E / k_B T
        + \\kappa N ( c_1 + \\phi_1 / 2 )^2 \\Big],

    where :math:`c_1` represents the concentration of species 1, i.e.
    :math:`c_1=N_1/N`. (Please note that the quantities :math:`\\kappa` and
    :math:`\\phi` correspond, respectively, to :math:`\\bar{\\kappa}` and
    :math:`\\bar{\\phi}` in [SadErh12]_.) This implementation requires
    :math:`\\phi` to be specified for both species. The sum of the specified
    :math:`\\phi` values is required to be :math:`-2`, because then the above
    expression is symmetric with respect to interchange of species 1 and 2,
    i.e., it does not matter if we use :math:`\\phi_1` and :math:`c_1` or
    :math:`\\phi_2` and :math:`c_2`.

    Just like the :ref:`semi-grand canonical ensemble <canonical_ensemble>`,
    the VCSGC ensemble allows concentrations to change. A trial step consists
    of changing the identity of a randomly chosen atom and accepting the change
    with probability

    .. math::

        P = \\min \\{ 1, \\, \\exp [ - \\Delta E / k_B T
        + \\kappa N \\Delta c_1 (\\phi_1 + \\Delta c_1 + 2 c_1 ) ] \\}.

    Note that for a sufficiently large value of :math:`\\kappa`, say 200, the
    probability density :math:`\\rho_{\\text{VCSGC}}` is sharply peaked around
    :math:`c_1=-\\phi_1 / 2`. In practice, this means that we can gradually
    change :math:`\\phi_1` from (using some margins) :math:`-2.1` to
    :math:`0.1` and take the system continuously from :math:`c_1 = 0` to
    :math:`1`. The parameter :math:`\\kappa` constrains the fluctuations (or
    the variance) of the concentration at each value of :math:`\\phi_1`, with
    higher values of :math:`\\kappa` meaning less fluctuations. Unlike the
    :ref:`semi-grand canonical ensemble <vcsgc_ensemble>`, one value of
    :math:`\\phi_1` maps to one and only one concentration also in multiphase
    regions. Since the derivative of the canonical free energy can be expressed
    in terms of parameters and observables of the VCSGC ensemble,

    .. math::

        k_B T \\kappa ( \\phi_1 + 2 \\langle c_1 \\rangle ) = - \\frac{1}{N}
        \\frac{\\partial F}{\\partial c_1} (N, V, T, \\langle c_1 \\rangle ),

    this ensemble allows for thermodynamic integration across multiphase
    regions. This means that we can construct phase diagrams by directly
    comparing the free energies of the different phases. This often makes the
    VCSGC ensemble more convenient than the :ref:`semi-grand canonical ensemble
    <sgc_ensemble>` when simulating materials with multiphase regions, such as
    alloys with miscibility gaps.

    When using the VCSGC ensemble, please cite Sadigh, B. and Erhart, P., Phys.
    Rev. B **86**, 134204 (2012) [SadErh12]_.

    Parameters
    ----------
    atoms : :class:`ase:Atoms`
        atomic configuration to be used in the Monte Carlo simulation;
        also defines the initial occupation vector
    calculator : :class:`BaseCalculator`
        calculator to be used for calculating the potential changes
        that enter the evaluation of the Metropolis criterion
    temperature : float
        temperature :math:`T` in appropriate units [commonly Kelvin]
    phis : Dict[str, float]
        average constraint parameters :math:`\\phi_i`; the key denotes the
        species; there must be one entry for each species but their sum must be
        :math:`-2.0` (referred to as :math:`\\bar{\\phi}` in [SadErh12]_)
    kappa : float
        parameter that constrains the variance of the concentration
        (referred to as :math:`\\bar{\\kappa}` in [SadErh12]_)
    boltzmann_constant : float
        Boltzmann constant :math:`k_B` in appropriate
        units, i.e. units that are consistent
        with the underlying cluster expansion
        and the temperature units [default: eV/K]
    user_tag : str
        human-readable tag for ensemble [default: None]
    data_container : str
        name of file the data container associated with the ensemble
        will be written to; if the file exists it will be read, the
        data container will be appended, and the file will be
        updated/overwritten
    random_seed : int
        seed for the random number generator used in the Monte Carlo
        simulation
    ensemble_data_write_interval : int
        interval at which data is written to the data container; this
        includes for example the current value of the calculator
        (i.e. usually the energy) as well as ensembles specific fields
        such as temperature or the number of atoms of different species
    data_container_write_period : float
        period in units of seconds at which the data container is
        written to file; writing periodically to file provides both
        a way to examine the progress of the simulation and to back up
        the data [default: np.inf]
    trajectory_write_interval : int
        interval at which the current occupation vector of the atomic
        configuration is written to the data container.
    """

    def __init__(self, atoms: Atoms, calculator: BaseCalculator,
                 temperature: float, phis: Dict[str, float],
                 kappa: float, boltzmann_constant: float = kB,
                 user_tag: str = None,
                 data_container: DataContainer = None,
                 random_seed: int = None,
                 data_container_write_period: float = np.inf,
                 ensemble_data_write_interval: int = None,
                 trajectory_write_interval: int = None) -> None:

        self._ensemble_parameters = dict(temperature=temperature,
                                         kappa=kappa)
        self._set_phis(phis)
        for atnum, phi in self.phis.items():
            phi_sym = 'phi_{}'.format(chemical_symbols[atnum])
            self._ensemble_parameters[phi_sym] = phi

        self._boltzmann_constant = boltzmann_constant

        super().__init__(
            atoms=atoms, calculator=calculator, user_tag=user_tag,
            data_container=data_container,
            random_seed=random_seed,
            data_container_write_period=data_container_write_period,
            ensemble_data_write_interval=ensemble_data_write_interval,
            trajectory_write_interval=trajectory_write_interval)

        if len(self.configuration._allowed_species) > 2:
            raise NotImplementedError('VCSGCEnsemble does not yet support '
                                      'cluster spaces with more than two '
                                      'species.')

        if set(self.configuration._allowed_species) != set(self.phis.keys()):
            raise ValueError('phis were not set for all species')

    def _do_trial_step(self):
        """ Carries out one Monte Carlo trial step. """
        self._total_trials += 1

        # choose flip
        sublattice_index = self.get_random_sublattice_index()
        index, new_species = \
            self.configuration.get_flip_state(sublattice_index)
        old_species = self.configuration.occupations[index]

        # Calculate difference in VCSGC thermodynamic potential.
        # Note that this assumes that only one atom was flipped.
        N = len(self.atoms)
        occupations = self.configuration._occupations.tolist()
        potential_diff = 1.0  # dN
        potential_diff -= occupations.count(old_species)
        potential_diff -= 0.5 * N * self.phis[old_species]
        potential_diff += occupations.count(new_species)
        potential_diff += 0.5 * N * self.phis[new_species]
        potential_diff *= self.kappa
        potential_diff *= self.boltzmann_constant * self.temperature
        potential_diff /= N

        potential_diff += self._get_property_change([index], [new_species])

        if self._acceptance_condition(potential_diff):
            self._accepted_trials += 1
            self.update_occupations([index], [new_species])

    def _acceptance_condition(self, potential_diff: float) -> bool:
        """
        Evaluates Metropolis acceptance criterion.

        Parameters
        ----------
        potential_diff
            the change in the thermodynamic potential associated
            with the trial step
        """
        if potential_diff < 0:
            return True
        else:
            return np.exp(-potential_diff / (
                self.boltzmann_constant * self.temperature)) > \
                self._next_random_number()

    @property
    def temperature(self) -> float:
        """ temperature :math:`T` (see parameters section above) """
        return self.ensemble_parameters['temperature']

    @property
    def boltzmann_constant(self) -> float:
        """ Boltzmann constant :math:`k_B` (see parameters section above) """
        return self._boltzmann_constant

    @property
    def phis(self) -> Dict[int, float]:
        """
        phis :math:`\\phi_i`, one for each species but their sum must be
        :math:`-2.0` (referred to as :math:`\\bar{\\phi}` in [SadErh12]_)
        """
        return self._phis

    @property
    def kappa(self) -> float:
        """
        kappa :math:`\\bar{\\kappa}` constrain parameter
        (see parameters section above)
        """
        return self.ensemble_parameters['kappa']

    def _set_phis(self, phis: Dict[Union[int, str], float]):
        """ Sets values of phis."""
        if not isinstance(phis, dict):
            raise TypeError('phis has the wrong type: {}'.format(type(phis)))
        if abs(sum(phis.values()) + 2) > 1e-6:
            raise ValueError('The sum of all phis must equal to -2')

        self._phis = {}
        for key, phi in phis.items():
            if isinstance(key, str):
                atomic_number = atomic_numbers[key]
                self._phis[atomic_number] = phi
            elif isinstance(key, int):
                self._phis[key] = phi

    def _get_ensemble_data(self) -> Dict:
        """
        Returns a dict with the default data of the ensemble. This includes
        atom counts and free energy derivative.
        """
        data = super()._get_ensemble_data()

        # free energy derivative
        atnum_1 = min(self.phis.keys())
        concentration = self.configuration._occupations.tolist().count(
            atnum_1) / len(self.atoms)
        data['free_energy_derivative'] = self.kappa * \
            self.boltzmann_constant * self.temperature * \
            (- 2 * concentration - self.phis[atnum_1])

        # species counts
        atoms = self.configuration.atoms
        unique, counts = np.unique(atoms.numbers, return_counts=True)
        # TODO: avoid accessing a protected member of a client class
        for atnum in self.configuration._allowed_species:
            data['{}_count'.format(chemical_symbols[atnum])] = 0
        for atnum, count in zip(unique, counts):
            data['{}_count'.format(chemical_symbols[atnum])] = count

        return data



# -*- coding: utf-8 -*-

from .configuration_manager import ConfigurationManager
from .data_container import DataContainer

"""
mchammer - Monte Carlo simulation module
"""

__project__ = 'icet-mchammer'
__description__ = 'A Pythonic approach to cluster expansions'
__authors__ = ['Mattias Ångqvist',
               'William Armando Muñoz',
               'Thomas Holm Rod',
               'Paul Erhart']
__copyright__ = '2018'
__license__ = 'MIT'
__credits__ = ['Mattias Ångqvist',
               'William Armando Muñoz',
               'Thomas Holm Rod',
               'Paul Erhart']
__version__ = '0.1'
__all__ = ['ConfigurationManager',
           'DataContainer']
__maintainer__ = 'The icet developers team'
__maintainer_email__ = 'icet@materialsmodeling.org'
__status__ = 'alpha-version'
__url__ = 'http://icet.materialsmodeling.org/'



import itertools
import numpy as np
from scipy.interpolate import griddata
from scipy.spatial import ConvexHull as ConvexHullSciPy
from scipy.spatial.qhull import QhullError
from typing import List, Union


class ConvexHull:
    """This class provides functionality for extracting the convex hull
    of the (free) energy of mixing. It is based on the `convex hull
    calculator in SciPy
    <http://docs.scipy.org/doc/scipy-dev/reference/\
generated/scipy.spatial.ConvexHull.html>`_.

    Parameters
    ----------
    concentrations : list of floats / list of lists of floats
        concentrations for each structure listed as ``[[c1, c2], [c1, c2],
        ...]``; for binaries, in which case there is only one independent
        concentration, the format ``[c1, c2, c3, ...]`` works as well.
    energies : list of floats
        energy (or energy of mixing) for each structure

    Attributes
    ----------
    concentrations : NumPy array (N, dimensions)
        concentrations of the `N` structures on the convex hull
    energies : NumPy array
        energies of the `N` structures on the convex hull
    dimensions : int
        number of independent concentrations needed to specify a point in
        concentration space (1 for binaries, 2 for ternaries etc.)
    structures : list of int
        indices of structures that constitute the convex hull (indices are
        defined by the order of their concentrations and energies are fed when
        initializing the ConvexHull object)

    Examples
    --------
    A `ConvexHull` object is easily initialized by providing lists of
    concentrations and energies::

        hull = ConvexHull(data['concentration'], data['mixing_energy'])

    after which one can for example plot the data (assuming a
    matplotlib axis object ``ax``)::

        ax.plot(hull.concentrations, hull.energies)

    or extract structures at or close to the convex hull::

        low_energy_structures = hull.extract_low_energy_structures(
            data['concentration'], data['mixing_energy'],
            tolerance=0.005, structures=list_of_structures)

    A complete example can be found in the :ref:`basic tutorial
    <tutorial_enumerate_structures>`.

    """

    def __init__(self,
                 concentrations: Union[List[float], List[List[float]]],
                 energies: List[float]) -> None:
        assert len(concentrations) == len(energies)
        # Prepare data in format suitable for SciPy-ConvexHull
        concentrations = np.array(concentrations)
        energies = np.array(energies)
        points = np.column_stack((concentrations, energies))
        self.dimensions = len(points[0]) - 1

        # Construct convex hull
        hull = ConvexHullSciPy(points)

        # Collect convex hull points in handy arrays
        concentrations = []
        energies = []
        for vertex in hull.vertices:
            if self.dimensions == 1:
                concentrations.append(points[vertex][0])
            else:
                concentrations.append(points[vertex][0:-1])
            energies.append(points[vertex][-1])
        concentrations = np.array(concentrations)
        energies = np.array(energies)

        structures = hull.vertices
        # If there is just one independent concentration, we'd better sort
        # according to it
        if self.dimensions == 1:
            ces = list(zip(*sorted(zip(concentrations, energies, structures))))
            self.concentrations = np.array(ces[0])
            self.energies = np.array(ces[1])
            self.structures = np.array(ces[2])
        else:
            self.concentrations = concentrations
            self.energies = energies
            self.structures = structures

        # Remove points that are above the "pure components plane"
        self._remove_points_above_tie_plane()

    def _remove_points_above_tie_plane(self, tol: float = 1e-6) -> None:
        """
        Remove all points on the convex hull that correspond to maximum rather
        than minimum energy.

        Parameters
        ----------
        tol
            Tolerance for what energy constitutes a lower one.
        """

        # Identify the "complex concentration hull", i.e. the extremal
        # concentrations. In the simplest case, these should simply be the
        # pure components.
        if self.dimensions == 1:
            # Then the ConvexHullScipy function doesn't work, so we just pick
            # the indices of the lowest and highest concentrations.
            vertices = []
            vertices.append(np.argmin(self.concentrations))
            vertices.append(np.argmax(self.concentrations))
            vertices = np.array(vertices)
        else:
            concentration_hull = ConvexHullSciPy(self.concentrations)
            vertices = concentration_hull.vertices

        # Remove all points of the convex energy hull that have an energy that
        # is higher than what would be gotten with pure components at the same
        # concentration. These points are mathematically on the convex hull,
        # but in the physically uninteresting upper part, i.e. they maximize
        # rather than minimize energy.
        to_delete = []
        for i, concentration in enumerate(self.concentrations):
            # The points on the convex concentration hull should always be
            # included, so skip them.
            if i in vertices:
                continue

            # The energy obtained as a linear combination of concentrations on
            # the convex hull is the "z coordinate" of the position on a
            # (hyper)plane in the (number of independent concentrations +
            # 1)-dimensional (N-D) space. This plane is spanned by N points.
            # If there are more vertices on the convex hull, we need to loop
            # over all combinations of N vertices.
            for plane in itertools.combinations(vertices,
                                                min(len(vertices),
                                                    self.dimensions + 1)):
                # Calculate energy that would be gotten with pure components
                # with ascribed concentration.
                energy_pure = griddata(self.concentrations[np.array(plane)],
                                       self.energies[np.array(plane)],
                                       concentration,
                                       method='linear')

                # Prepare to delete if the energy was lowered. `griddata` gives
                # NaN if the concentration is outside the triangle formed by
                # the three vertices. The result of the below comparison is
                # then False, which is what we want.
                if energy_pure < self.energies[i] - tol:
                    to_delete.append(i)
                    break

        # Finally remove all points
        self.concentrations = np.delete(self.concentrations, to_delete, 0)
        self.energies = np.delete(self.energies, to_delete, 0)
        self.structures = list(np.delete(self.structures, to_delete, 0))

    def get_energy_at_convex_hull(self, target_concentrations:
                                  Union[List[float],
                                        List[List[float]]]) -> np.ndarray:
        """Returns the energy of the convex hull at specified concentrations.
        If any concentration is outside the allowed range, NaN is
        returned.

        Parameters
        ----------
        target_concentrations
            concentrations at target points

            If there is one independent concentration, a list of
            floats is sufficient. Otherwise, the concentrations ought
            to be provided as a list of lists, such as ``[[0.1, 0.2],
            [0.3, 0.1], ...]``.
        """
        if self.dimensions > 1:
            assert len(target_concentrations[0]) == self.dimensions

        # Loop over all complexes of N+1 points to make sure that the lowest
        # energy plane is used in the end. This is needed in two dimensions
        # but in higher.
        hull_candidate_energies = []
        for plane in itertools.combinations(range(len(self.energies)),
                                            min(len(self.energies),
                                                self.dimensions + 1)):
            try:
                plane_energies = griddata(self.concentrations[list(plane)],
                                          self.energies[list(plane)],
                                          np.array(target_concentrations),
                                          method='linear')
            except QhullError:
                # If the points lie on a line, the convex hull will fail, but
                # we do not need to care about these "planes" anyway
                continue
            hull_candidate_energies.append(plane_energies)

        # Pick out the lowest energies found
        hull_energies = np.nanmin(hull_candidate_energies, axis=0)
        return hull_energies

    def extract_low_energy_structures(self, concentrations:
                                      Union[List[float],
                                            List[List[float]]],
                                      energies: List[float],
                                      energy_tolerance: float, structures:
                                      list = None) -> List[int]:
        """Returns structures that lie within a certain tolerance of the convex
        hull.

        Parameters
        ----------
        concentrations
            concentrations of candidate structures

            If there is one independent concentration, a list of
            floats is sufficient. Otherwise, the concentrations must
            be provided as a list of lists, such as ``[[0.1, 0.2],
            [0.3, 0.1], ...]``.
        energies
            energies of candidate structures
        energy_tolerance
            include structures with an energy that is at most this far
            from the convex hull
        structures
            list of candidate structures, e.g., :class:`ASE Atoms
            <ase.Atoms>` objects, corresponding to ``concentrations``
            and ``energies``

            The list will be returned, but with the objects too
            far from the convex hull removed. If `None`, a list of
            indices is returned instead.

        """
        number_of_candidates = len(concentrations)
        assert len(energies) == number_of_candidates
        if structures is None:
            structures = list(range(number_of_candidates))
        assert len(structures) == number_of_candidates

        # Calculate energy at convex hull for specified concentrations
        hull_energies = self.get_energy_at_convex_hull(concentrations)

        # Check which ones were close enough
        close_to_hull_structures = []
        for i in range(number_of_candidates):
            if energies[i] <= hull_energies[i] + energy_tolerance:
                close_to_hull_structures.append(structures[i])

        return close_to_hull_structures



import os
import random
from abc import ABC, abstractmethod
from math import gcd
from time import time
from typing import Dict, List, Union, BinaryIO, TextIO

import numpy as np

from ase import Atoms
from ase.data import chemical_symbols
from collections import OrderedDict

from ..calculators.base_calculator import BaseCalculator
from ..configuration_manager import ConfigurationManager
from ..data_container import DataContainer
from ..observers.base_observer import BaseObserver


class BaseEnsemble(ABC):
    """Base ensemble class.

    Parameters
    ----------
    calculator : :class:`BaseCalculator`
        calculator to be used for calculating the potential changes
        that enter the evaluation of the Metropolis criterion
    atoms : :class:`ase:Atoms`
        atomic configuration to be used in the Monte Carlo simulation;
        also defines the initial occupation vector
    user_tag : str
        human-readable tag for ensemble [default: None]
    data_container : str
        name of file the data container associated with the ensemble
        will be written to; if the file exists it will be read, the
        data container will be appended, and the file will be
        updated/overwritten
    ensemble_data_write_interval : int
        interval at which data is written to the data container; this
        includes for example the current value of the calculator
        (i.e. usually the energy) as well as ensembles specific fields
        such as temperature or the number of atoms of different species
    data_container_write_period : float
        period in units of seconds at which the data container is
        written to file; writing periodically to file provides both
        a way to examine the progress of the simulation and to back up
        the data [default: np.inf]
    trajectory_write_interval : int
        interval at which the current occupation vector of the atomic
        configuration is written to the data container.
    random_seed : int
        seed for the random number generator used in the Monte Carlo
        simulation
    """

    def __init__(self, atoms: Atoms, calculator: BaseCalculator,
                 user_tag: str = None, data_container: DataContainer = None,
                 data_container_write_period: float = np.inf,
                 ensemble_data_write_interval: int = None,
                 trajectory_write_interval: int = None,
                 random_seed: int = None) -> None:

        # initialize basic variables
        self._accepted_trials = 0
        self._total_trials = 0
        self._observers = {}
        self._step = 0

        # calculator and configuration
        self._calculator = calculator
        self._user_tag = user_tag
        strict_constraints_symbol = self.calculator.occupation_constraints
        symbols = list({tuple(sym)
                        for sym in strict_constraints_symbol if len(sym) > 1})
        sublattices = [[] for _ in symbols]
        for i, constraint in enumerate(strict_constraints_symbol):
            for j, sym in enumerate(symbols):
                if len(sym) < 2:
                    continue
                if sorted(constraint) == sorted(sym):
                    sublattices[j].append(i)
        self._sublattices = sublattices
        strict_constraints = []
        for symbols in strict_constraints_symbol:
            numbers = []
            for symbol in symbols:
                numbers.append(chemical_symbols.index(symbol))
            strict_constraints.append(numbers)
        self.configuration = ConfigurationManager(
            atoms, strict_constraints, sublattices)

        # random number generator
        if random_seed is None:
            self._random_seed = random.randint(0, 1e16)
        else:
            self._random_seed = random_seed
        random.seed(a=self._random_seed)

        # add ensemble parameters and metadata
        self._ensemble_parameters['n_atoms'] = len(self.atoms)
        metadata = OrderedDict(ensemble_name=self.__class__.__name__,
                               user_tag=user_tag,
                               seed=self.random_seed)

        # data container
        self._data_container_write_period = data_container_write_period

        self._data_container_filename = data_container

        if data_container is not None and os.path.isfile(data_container):
            self._data_container = DataContainer.read(data_container)

            dc_ensemble_parameters = self.data_container.ensemble_parameters
            if self.ensemble_parameters != dc_ensemble_parameters:
                raise ValueError('Ensemble parameters do not match with those'
                                 ' stored in DataContainer file: {}'.format(
                                     set(dc_ensemble_parameters.items()) -
                                     set(self.ensemble_parameters.items())))
            self._restart_ensemble()
        else:
            if data_container is not None:
                # check if path to file exists
                filedir = os.path.dirname(data_container)
                if filedir and not os.path.isdir(filedir):
                    raise FileNotFoundError('Path to data container file does'
                                            ' not exist: {}'.format(filedir))
            self._data_container = \
                DataContainer(atoms=atoms,
                              ensemble_parameters=self.ensemble_parameters,
                              metadata=metadata)

        # interval for writing data and further preparation of data container
        default_interval = max(1, 10 * round(len(atoms) / 10))

        if ensemble_data_write_interval is None:
            self._ensemble_data_write_interval = default_interval
        else:
            self._ensemble_data_write_interval = ensemble_data_write_interval

        # Handle trajectory writing
        if trajectory_write_interval is None:
            self._trajectory_write_interval = default_interval
        else:
            self._trajectory_write_interval = trajectory_write_interval

        self._find_observer_interval()

    @property
    def atoms(self) -> Atoms:
        """ current configuration (copy) """
        return self.configuration.atoms.copy()

    @property
    def total_trials(self) -> int:
        """ number of Monte Carlo trial steps """
        return self._total_trials

    @property
    def accepted_trials(self) -> int:
        """ number of accepted trial steps """
        return self._accepted_trials

    @property
    def data_container(self) -> DataContainer:
        """ data container associated with ensemble """
        return self._data_container

    @property
    def observers(self) -> Dict[str, BaseObserver]:
        """ observers """
        return self._observers

    @property
    def acceptance_ratio(self) -> float:
        """ acceptance ratio """
        if self.total_trials > 0:
            return self.accepted_trials / self.total_trials
        return 0

    @property
    def calculator(self) -> BaseCalculator:
        """ calculator attached to the ensemble """
        return self._calculator

    @property
    def data_container_write_period(self) -> float:
        " data container write period "
        return self._data_container_write_period

    @data_container_write_period.setter
    def data_container_write_period(self, data_container_write_period):
        self._data_container_write_period = data_container_write_period

    def run(self, number_of_trial_steps: int, reset_step: bool = False):
        """
        Samples the ensemble for the given number of trial steps.

        Parameters
        ----------
        number_of_trial_steps
            number of MC trial steps to run in total
        reset_step
            if True the MC trial step counter and the data container will
            be reset to zero and empty, respectively.
        """

        last_write_time = time()
        if reset_step:
            initial_step = 0
            final_step = number_of_trial_steps
            self.reset_data_container()
        else:
            initial_step = self._step
            final_step = self._step + number_of_trial_steps
            # run Monte Carlo simulation such that we start at an
            # interval which lands on the observer interval
            if not initial_step == 0:
                first_run_interval = self.observer_interval -\
                    (initial_step -
                     (initial_step // self.observer_interval) *
                        self.observer_interval)
                first_run_interval = min(
                    first_run_interval, number_of_trial_steps)
                self._run(first_run_interval)
                initial_step += first_run_interval
                self._step += first_run_interval

        step = initial_step
        while step < final_step:
            uninterrupted_steps = min(
                self.observer_interval, final_step - step)
            if self._step % self.observer_interval == 0:
                self._observe(self._step)
            if self._data_container_filename is not None and \
                    time() - last_write_time > \
                    self.data_container_write_period:
                self.write_data_container(self._data_container_filename)
                last_write_time = time()

            self._run(uninterrupted_steps)
            step += uninterrupted_steps
            self._step += uninterrupted_steps

        # If we end on an observation interval we also observe
        if self._step % self.observer_interval == 0:
            self._observe(self._step)

        if self._data_container_filename is not None:
            self.write_data_container(self._data_container_filename)

    def _run(self, number_of_trial_steps: int):
        """Runs MC simulation for a number of trial steps without
        interruption.

        Parameters
        ----------
        number_of_trial_steps
            number of trial steps to run without stopping
        """
        for _ in range(number_of_trial_steps):
            self._do_trial_step()

    def _observe(self, step: int):
        """Submits current configuration to observers and appends
        observations to data container.

        Parameters
        ----------
        step
            the current trial step
        """
        row_dict = {}

        # Ensemble specific data
        if step % self._ensemble_data_write_interval == 0:
            ensemble_data = self._get_ensemble_data()
            for key, value in ensemble_data.items():
                row_dict[key] = value

        # Trajectory data
        if step % self._trajectory_write_interval == 0:
            row_dict['occupations'] = self.configuration.occupations.tolist()

        # Observer data
        for observer in self.observers.values():
            if step % observer.interval == 0:
                if observer.return_type is dict:
                    for key, value in observer.get_observable(
                            self.calculator.atoms).items():
                        row_dict[key] = value
                else:
                    row_dict[observer.tag] = observer.get_observable(
                        self.calculator.atoms)

        if len(row_dict) > 0:
            self._data_container.append(mctrial=step, record=row_dict)

    @abstractmethod
    def _do_trial_step(self):
        pass

    @property
    def user_tag(self) -> str:
        """ tag used for labeling the ensemble """
        return self._user_tag

    @property
    def random_seed(self) -> int:
        """ seed used to initialize random number generator """
        return self._random_seed

    def _next_random_number(self) -> float:
        """ Returns the next random number from the PRNG. """
        return random.random()

    @property
    def observer_interval(self) -> int:
        """minimum number of steps to run Monte Carlo simulation without
        interruption for observation
        """
        return self._observer_interval

    def _find_observer_interval(self) -> int:
        """
        Finds the greatest common denominator from the observation intervals.
        """
        intervals = [obs.interval for obs in self.observers.values()]

        if self._ensemble_data_write_interval is not np.inf:
            intervals.append(self._ensemble_data_write_interval)
        if self._trajectory_write_interval is not np.inf:
            intervals.append(self._trajectory_write_interval)
        if intervals:
            self._observer_interval = self._get_gcd(intervals)

    def _get_gcd(self, values: List[int]) -> int:
        """
        Finds the greatest common denominator (GCD) from a list.
        """
        if len(values) == 1:
            return values[0]

        if len(values) > 2:
            gcd_right = gcd(values[-1], values[-2])
            values.pop()
            values.pop()
            values.append(gcd_right)
            return self._get_gcd(values)
        else:
            return gcd(values[0], values[1])

    def attach_observer(self, observer: BaseObserver, tag=None):
        """
        Attaches an observer to the ensemble.

        Parameters
        ----------
        observer
            observer instance to attach
        tag
            name used in data container
        """
        if not isinstance(observer, BaseObserver):
            raise TypeError('observer has the wrong type: {}'
                            .format(type(observer)))

        if tag is not None:
            observer.tag = tag
            self.observers[tag] = observer
        else:
            self.observers[observer.tag] = observer

        self._find_observer_interval()

    def reset_data_container(self):
        """ Resets the data container and the trial step counter. """
        self._step = 0
        self._total_trials = 0
        self._accepted_trials = 0

        self._data_container.reset()

    def update_occupations(self, sites: List[int], species: List[int]):
        """Updates the occupation vector of the configuration being sampled.
        This will change the state of the configuration in both the
        calculator and the configuration manager.

        Parameters
        ----------
        sites
            indices of sites of the configuration to change
        species
            new occupations (species) by atomic number

        Raises
        ------
        ValueError
            if input lists are not of the same length
        """

        if len(sites) != len(species):
            raise ValueError('sites and species must have the same length.')
        self.configuration.update_occupations(sites, species)

    def _get_property_change(self,
                             sites: List[int], species: List[int]) -> float:
        """Computes and returns the property change due to a change of the
        configuration.

        _N.B.:_ This method leaves to configuration itself unchanged.

        Parameters
        ----------
        sites
            indices of sites to change
        species
            new occupations (species) by atomic number
        """
        current_species = self.configuration.occupations[sites]
        current_property = self.calculator.calculate_local_contribution(
            local_indices=sites,
            occupations=self.configuration.occupations)

        self.update_occupations(sites=sites, species=species)
        new_property = self.calculator.calculate_local_contribution(
            local_indices=sites,
            occupations=self.configuration.occupations)
        property_change = new_property - current_property

        # Restore initial configuration
        self.update_occupations(sites, current_species)
        return property_change

    def _get_ensemble_data(self) -> dict:
        """ Returns the current calculator property. """
        return {
            'potential': self.calculator.calculate_total(
                occupations=self.configuration.occupations),
            'acceptance_ratio': self.acceptance_ratio}

    def get_random_sublattice_index(self) -> int:
        """Returns a random sublattice index based on the weights of the
        sublattice.

        Todo
        ----
        * fix this method
        * add unit test
        """
        total_active_sites = sum([len(sub) for sub in self._sublattices])
        probability_distribution = [
            len(sub) / total_active_sites for sub in self._sublattices]
        pick = np.random.choice(
            range(0, len(self._sublattices)), p=probability_distribution)
        return pick

    def _restart_ensemble(self):
        """Restarts ensemble using the last state saved in DataContainer file.
        """

        # Restart step
        self._step = self.data_container.last_state['last_step']

        # Update configuration
        occupations = self.data_container.last_state['occupations']
        sites = list(range(len(self.configuration.atoms)))
        self.update_occupations(sites, occupations)

        # Restart number of total and accepted trial steps
        self._total_trials = self._step
        self._accepted_trials = \
            self.data_container.last_state['accepted_trials']

        # Restart state of random number generator
        random.setstate(self.data_container.last_state['random_state'])

    def write_data_container(self, outfile: Union[str, BinaryIO, TextIO]):
        """Updates last state of the Monte Carlo simulation and
        writes DataContainer to file.

        Parameters
        ----------
        outfile
            file to which to write
        """
        self._data_container._update_last_state(
            last_step=self._step,
            occupations=self.configuration.occupations.tolist(),
            accepted_trials=self._accepted_trials,
            random_state=random.getstate())

        self.data_container._write(outfile)

    @property
    def ensemble_parameters(self) -> dict:
        """Returns parameters associated with the ensemble."""
        return self._ensemble_parameters.copy()



import numpy as np
import spglib
from ase import Atoms
from typing import Tuple, List, Sequence, TypeVar

from ase.neighborlist import NeighborList as ase_NeighborList
from icet.core.neighbor_list import NeighborList
from icet.core.structure import Structure
from icet.core.lattice_site import LatticeSite
from ase.data import chemical_symbols

Vector = List[float]
T = TypeVar('T')


def get_scaled_positions(positions: np.ndarray,
                         cell: np.ndarray,
                         wrap: bool = True,
                         pbc: List[bool] = [True, True, True]) \
        -> np.ndarray:
    """
    Returns the positions in reduced (or scaled) coordinates.

    Parameters
    ----------
    positions
        atomic positions in cartesian coordinates
    cell
        cell metric
    wrap
        if True, positions outside the unit cell will be wrapped into
        the cell in the directions with periodic boundary conditions
        such that the scaled coordinates are between zero and one.
    pbc
        periodic boundary conditions flags
    """

    fractional = np.linalg.solve(cell.T, positions.T).T

    if wrap:
        for i, periodic in enumerate(pbc):
            if periodic:
                # Yes, we need to do it twice.
                # See the scaled_positions.py test.
                fractional[:, i] %= 1.0
                fractional[:, i] %= 1.0

    return fractional


def add_vacuum_in_non_pbc(configuration: Atoms) -> Atoms:
    """
    Adds vacuum in non-periodic directions.

    Parameters
    ----------
    configuration
        input atomic structure

    Returns
    -------
    ase.Atoms
        input atomic structure with vacuum in non-pbc directions
    """
    configuration_cpy = configuration.copy()
    vacuum_axis = []
    for i, pbc in enumerate(configuration.pbc):
        if not pbc:
            vacuum_axis.append(i)

    if len(vacuum_axis) > 0:
        configuration_cpy.center(30, axis=vacuum_axis)
        configuration_cpy.wrap()

    return configuration_cpy


def get_primitive_structure(atoms: Atoms, no_idealize: bool = True,
                            to_primitive=True, symprec=1e-5) -> Atoms:
    """
    Determines primitive structure using spglib.

    Parameters
    ----------
    atoms
        input atomic structure
    no_idealize
        If True lengths and angles are not idealized

    Returns
    -------
    atoms_prim
        primitive structure
    """
    atoms_cpy = atoms.copy()
    atoms_with_vacuum = add_vacuum_in_non_pbc(atoms_cpy)

    atoms_as_tuple = ase_atoms_to_spglib_cell(atoms_with_vacuum)

    lattice, scaled_positions, numbers = spglib.standardize_cell(
        atoms_as_tuple, to_primitive=to_primitive,
        no_idealize=no_idealize, symprec=symprec)
    scaled_positions = [np.round(pos, 12) for pos in scaled_positions]
    atoms_prim = Atoms(scaled_positions=scaled_positions,
                       numbers=numbers, cell=lattice, pbc=atoms.pbc)
    atoms_prim.wrap()

    return atoms_prim


def get_fractional_positions_from_neighbor_list(
        structure: Structure, neighbor_list: NeighborList) -> List[Vector]:
    """
    Returns the fractional positions of the lattice sites in structure from
    a neighbor list.

    Parameters
    ----------
    atoms
        input atomic structure
    neighbor_list
        list of lattice neighbors of the input structure
    """
    neighbor_positions = []
    fractional_positions = []
    lattice_site = LatticeSite(0, [0, 0, 0])

    for i in range(len(neighbor_list)):
        lattice_site.index = i
        position = structure.get_position(lattice_site)
        neighbor_positions.append(position)
        for neighbor in neighbor_list.get_neighbors(i):
            position = structure.get_position(neighbor)
            neighbor_positions.append(position)

    if len(neighbor_positions) > 0:
        fractional_positions = get_scaled_positions(
            np.array(neighbor_positions),
            structure.cell, wrap=False,
            pbc=structure.pbc)

    return fractional_positions


def get_fractional_positions_from_ase_neighbor_list(
        atoms: Atoms, neighbor_list: ase_NeighborList) -> List[Vector]:
    """
    Returns the fractional positions of the lattice sites in atomic structure
    from a neighbor list.

    Parameters
    ----------
    atoms
        input atomic structure
    neighbor_list
        list of neighbors of the input structure
    """
    neighbor_positions = []
    fractional_positions = []

    for i in range(len(atoms)):
        lattice_site = LatticeSite(i, [0., 0., 0.])
        position = get_position_from_lattice_site(atoms, lattice_site)
        neighbor_positions.append(position)
        indices, offsets = neighbor_list.get_neighbors(i)
        for index, offset in zip(indices, offsets):
            lattice_site = LatticeSite(index, offset)
            position = get_position_from_lattice_site(atoms, lattice_site)
            neighbor_positions.append(position)
    if len(neighbor_positions) > 0:
        fractional_positions = get_scaled_positions(
            np.array(neighbor_positions),
            atoms.cell, wrap=False,
            pbc=atoms.pbc)
    return fractional_positions


def get_position_from_lattice_site(atoms: Atoms, lattice_site: LatticeSite):
    """
    Gets the corresponding position from the lattice site.

    Parameters
    ---------
    atoms
        input atomic structure
    lattice_site
        specific lattice site of the input structure
    """
    return atoms[lattice_site.index].position + \
        np.dot(lattice_site.unitcell_offset, atoms.get_cell())


def find_lattice_site_by_position(atoms: Atoms, position: List[float],
                                  tol: float = 1e-4) -> LatticeSite:
    """
    Tries to construct a lattice site equivalent from
    position in reference to the atoms object.

    atoms
        input atomic structure
    position
        presumed cartesian coordinates of a lattice site
    """

    for i, atom in enumerate(atoms):
        pos = position - atom.position
        # Direct match
        if np.linalg.norm(pos) < tol:
            return LatticeSite(i, np.array((0., 0., 0.)))

        fractional = np.linalg.solve(atoms.cell.T, np.array(pos).T).T
        unit_cell_offset = [np.floor(round(x)) for x in fractional]
        residual = np.dot(fractional - unit_cell_offset, atoms.cell)
        if np.linalg.norm(residual) < tol:
            latNbr = LatticeSite(i, unit_cell_offset)
            return latNbr

    # found nothing, raise error
    raise RuntimeError('Did not find site in find_lattice_site_by_position')


def fractional_to_cartesian(atoms: Atoms,
                            frac_positions: List[Vector]) -> List[Vector]:
    """
    Turns fractional positions into cartesian positions.

    Parameters
    ----------
    atoms
        input atomic structure
    frac_positions
        fractional positions
    """
    return np.dot(frac_positions, atoms.cell)


def get_permutation(container: Sequence[T],
                    permutation: List[int]) -> Sequence[T]:
    """
    Returns the permuted version of container.
    """
    if len(permutation) != len(container):
        raise RuntimeError('Container and permutation'
                           ' not of same size {} != {}'.format(
                               len(container), len(permutation)))
    if len(set(permutation)) != len(permutation):
        raise Exception
    return [container[s] for s in permutation]


def find_permutation(target: Sequence[T],
                     permutated: Sequence[T]) -> List[int]:
    """
    Returns the permutation vector that takes permutated to target.

    Parameters
    ----------
    target
        a container that supports indexing and its elements contain
        objects with __eq__ method
    permutated
        a container that supports indexing and its elements contain
        objects with __eq__ method
    """
    permutation = []
    for element in target:
        index = permutated.index(element)
        permutation.append(index)
    return permutation


def ase_atoms_to_spglib_cell(atoms: Atoms) \
        -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Returns a tuple of three components: cell metric, atomic positions, and
    atomic species of the input ASE Atoms object.
    """
    return (atoms.get_cell(),
            atoms.get_scaled_positions(),
            atoms.get_atomic_numbers())


def get_decorated_primitive_structure(
        atoms: Atoms, allowed_species: List[List[str]])-> Tuple[
        Atoms, List[List[str]]]:
    """Returns a decorated primitive structure

    Example
    --------
        Will put hydrogen on sublattice 1, Helium on sublattice 2 and
        so on

    todo : simplify the revert back to unsorted symbols
    """
    if len(atoms) != len(allowed_species):
        raise ValueError(
            'Atoms object and chemical symbols need to be the same size.')
    symbols = set()
    symbols = sorted({tuple(sorted(s)) for s in allowed_species})

    decorated_primitive = atoms.copy()
    for i, sym in enumerate(allowed_species):
        sublattice = symbols.index(tuple(sorted(sym))) + 1
        decorated_primitive[i].symbol = chemical_symbols[sublattice]

    decorated_primitive = get_primitive_structure(decorated_primitive)
    decorated_primitive.wrap()
    primitive_chemical_symbols = []
    for atom in decorated_primitive:
        sublattice = chemical_symbols.index(atom.symbol)
        primitive_chemical_symbols.append(symbols[sublattice-1])

    for symbols in allowed_species:
        if tuple(sorted(symbols)) in primitive_chemical_symbols:
            index = primitive_chemical_symbols.index(tuple(sorted(symbols)))
            primitive_chemical_symbols[index] = symbols
    return decorated_primitive, primitive_chemical_symbols



import logging
from typing import List, Tuple

import numpy as np
from ase import Atoms
from ase.build import cut

from icet.io.logging import logger
logger = logger.getChild('structure_mapping')


def map_structure_to_reference(input_structure: Atoms,
                               reference_structure: Atoms,
                               tolerance_mapping: float,
                               vacancy_type: str = None,
                               inert_species: List[str] = None,
                               tolerance_cell: float = 0.05,
                               tolerance_positions: float = 0.01) \
                               -> Tuple[Atoms, float, float]:
    """Maps a relaxed structure onto a reference structure.
    The function returns a tuple comprising

    * the ideal supercell most closely matching the input structure,
    * the largest deviation of any input coordinate from its ideal
      coordinate, and
    * the average deviation of the input coordinates from the ideal
      coordinates.

    Parameters
    ----------
    input_structure
        relaxed input structure
    reference_structure
        reference structure, which can but need not represent the primitive
        cell
    tolerance_mapping
        maximum allowed displacement for mapping an atom in the relaxed (but
        rescaled) structure to the reference supercell

        *Note*: A reasonable choice is up to 20-30% of the first
        nearest neighbor distance (`r1`).  A value above 50% of `r1`
        will most likely lead to atoms being multiply assigned,
        whereby the mapping fails.
    vacancy_type
        If this parameter is set to a non-zero string unassigned sites in the
        reference structure will be assigned to this type.

        *Note 1*: By default (``None``) the method will fail if there
        are *any* unassigned sites in the reference structure.

        *Note 2*: ``vacancy_type`` must be a valid species as
        enforced by the :class:`ase.Atoms` class.
    inert_species
        List of chemical symbols (e.g., ``['Au', 'Pd']``) that are never
        substituted for a vacancy. Used to make an initial rescale of the cell
        and thus increases the probability for a successful mapping. Need not
        be specified if ``vacancy_type`` is ``None``.
    tolerance_cell
        tolerance factor applied when computing permutation matrix to generate
        supercell
    tolerance_positions
        tolerance factor applied when scanning for overlapping positions in
        Angstrom (forwarded to :func:`ase.build.cut`)

    Example
    -------
    The following code snippet illustrates the general usage. It first creates
    a primitive FCC cell, which is latter used as reference structure. To
    emulate a relaxed structure obtained from, e.g., a density functional
    theory calculation, the code then creates a 4x4x4 conventional FCC
    supercell, which is populated with two different atom types, has distorted
    cell vectors, and random displacements to the atoms. Finally, the present
    function is used to map the structure back the ideal lattice::

        from ase.build import bulk
        reference = bulk('Au', a=4.09)
        atoms = bulk('Au', cubic=True, a=4.09).repeat(4)
        atoms.set_chemical_symbols(10 * ['Ag'] + (len(atoms) - 10) * ['Au'])
        atoms.set_cell(atoms.cell * 1.02, scale_atoms=True)
        atoms.rattle(0.1)
        mapped_atoms = map_structure_to_reference(atoms, reference, 1.0)

    """
    assert np.all(input_structure.pbc == reference_structure.pbc), \
        ('The periodic boundary conditions differ'
         ' between input and reference structure')

    # Scale input cell and construct supercell of the reference structure
    scaled_cell = _get_scaled_cell(input_structure, reference_structure,
                                   vacancy_type=vacancy_type,
                                   inert_species=inert_species)
    P = _get_transformation_matrix(scaled_cell,
                                   reference_structure.cell,
                                   tolerance_cell=tolerance_cell)
    scaled_structure, ideal_supercell = \
        _rescale_structures(input_structure,
                            reference_structure,
                            P,
                            tolerance_positions=tolerance_positions)

    assert len(ideal_supercell) == len(scaled_structure) or \
        vacancy_type is not None, \
        ('Number of atoms in ideal supercell does not match '
         'input structure.\n'
         'ideal: {}\ninput: {}'.format(len(ideal_supercell),
                                       len(scaled_structure)))

    if logger.isEnabledFor(logging.DEBUG):
        np.set_printoptions(suppress=True, precision=6)
        logger.debug('Number of atoms in reference structure:'
                     ' {}'.format(len(reference_structure)))
        logger.debug('Number of atoms in input structure:'
                     ' {}\n'.format(len(input_structure)))
        logger.debug('Reference cell metric:\n'
                     '{}'.format(reference_structure.cell))
        logger.debug('Input cell metric:\n'
                     '{}\n'.format(input_structure.cell))
        logger.debug('Transformation matrix connecting reference structure'
                     ' and idealized input structure:\n {}'.format(P))
        logger.debug('Determinant of transformation matrix:'
                     ' {:.3f}\n'.format(np.linalg.det(P)))
        logger.debug('Cell metric of ideal supercell:\n'
                     '{}'.format(ideal_supercell.cell))
        logger.debug('Cell metric of rescaled input structure:\n'
                     '{}\n'.format(scaled_structure.cell))

    # map atoms in input structure to closest site in ideal supercell
    dr_max = 0.0
    dr_sum = 0.0
    dr_sumsq = 0.0
    # per-atom-list for keeping track of mapped atoms
    mapped = [-1] * len(ideal_supercell)
    # distances between ideal and input sites
    drs = [None] * len(ideal_supercell)
    for ideal_site in ideal_supercell:
        for atom in scaled_structure:
            # in order to compute the distance the current atom from
            # the input structure is temporarily added to the
            # ideal supercell. This allows one to simply use the ASE
            # Atoms method for computing the interatomic distance
            ideal_supercell.append(atom)
            dr = ideal_supercell.get_distance(ideal_site.index,
                                              ideal_supercell[-1].index,
                                              mic=True)
            del ideal_supercell[-1]
            if dr < tolerance_mapping:
                assert mapped[ideal_site.index] < 0, \
                    ('More than one atom from the relaxed'
                     ' (and rescaled) structure have been'
                     ' mapped onto the same ideal site.\n'
                     ' Try reducing `tolerance_mapping`.')
                mapped[ideal_site.index] = atom.index
                drs[ideal_site.index] = dr
                ideal_site.symbol = atom.symbol
                dr_max = max(dr, dr_max)
                dr_sum += dr
                dr_sumsq += dr * dr
                break
        else:
            assert vacancy_type is not None, \
                ('Failed to assign an atom from the relaxed (and'
                 ' rescaled) structure to the ideal lattice.'
                 ' Try increasing `tolerance_mapping`.\n'
                 ' {}'.format(ideal_site))
            ideal_site.symbol = vacancy_type

    dr_avg = dr_sum / len(ideal_supercell)
    dr_sdv = np.sqrt(dr_sumsq / len(ideal_supercell) - dr_avg ** 2)

    # Check that not more than one atom was assigned to the same site
    for k in set(mapped):
        assert k < 0 or mapped.count(k) <= 1, \
            ('Site {} has been assigned more than once.'.format(k))

    # Check that the chemical composition of input and ideal supercell matches
    for symbol in set(input_structure.get_chemical_symbols()):
        n1 = input_structure.get_chemical_symbols().count(symbol)
        n2 = ideal_supercell.get_chemical_symbols().count(symbol)
        assert n1 == n2, ('Number of atoms of type {} differs between'
                          ' input structure ({}) and ideal'
                          ' supercell ({}).'.format(symbol, n1, n2))

    if logger.isEnabledFor(logging.DEBUG):
        logger.debug('Maximum, average and standard deviation of atomic'
                     ' displacements: {} {} {}'.format(dr_max, dr_avg, dr_sdv))

        logger.debug('{:52} {:}'.format('Input structure:',
                                        'Scaled structure:'))
        for k, (input_atom, scaled_atom) in enumerate(zip(input_structure,
                                                          scaled_structure)):
            msg = '{:4}  {:2}'.format(k, input_atom.symbol)
            msg += (3 * ' {:12.6f}').format(*input_atom.position)
            msg += '    -->'
            msg += (3 * ' {:12.6f}').format(*scaled_atom.position)
            logger.debug(msg)

        logger.debug('\n')

        logger.debug('{:52} {}'.format('Ideal supercell:',
                                       'Scaled structure:'))
        for ideal_atom, k, dr in zip(ideal_supercell, mapped, drs):
            msg = ' {:2}'.format(ideal_atom.symbol)
            msg += (3 * '  {:12.6f}').format(*ideal_atom.position)
            msg += '    -->'
            msg += ' {:4}'.format(k)

            if k >= 0:
                scaled_pos = scaled_structure[k].position
                msg += (3 * ' {:12.6f}').format(*scaled_pos)
                msg += '    --> {:.4f}'.format(dr)
            logger.debug(msg)

    return ideal_supercell, dr_max, dr_avg


def _get_scaled_cell(input_structure: Atoms,
                     reference_structure: Atoms,
                     vacancy_type: str = None,
                     inert_species: List[str] = None) -> np.ndarray:
    """
    The input structure needs to be scaled in order to match the lattice
    structure of the reference structure. The reference structure can be a
    primitive cell, in which case the input structure would usually be a
    supercell thereof. Also, we need an ideal supercell that matches the input
    structure.

    Parameters
    - ---------
    input_structure
        relaxed input structure
    reference_structure: ASE Atoms object
        reference structure, which can but need not represent the primitive
        cell
    vacancy_type
        If not None, the cell is scaled if and only if `inert_species` is not
        None
    inert_species
        List of chemical symbols(e.g., `['Au', 'Pd']`) that are never
        substituted for a vacancy. Needless if `vacancy_type` is `None`
    """
    modcell = input_structure.get_cell()
    if vacancy_type is None:
        # Without scale factor we can just rescale with number of atoms
        atvol_in = input_structure.get_volume() / len(input_structure)
        atvol_ref = reference_structure.get_volume() / len(reference_structure)
        scale = atvol_in / atvol_ref
    if vacancy_type is not None:
        if inert_species is None:
            scale = 1.0
        else:
            # We can not use the number of atoms since there may be vacancies
            # in the input_structure. Instead we count the species that we
            # know should always be present.
            n_in = 0
            n_ref = 0
            symbols_in = input_structure.get_chemical_symbols()
            symbols_ref = reference_structure.get_chemical_symbols()
            for species in inert_species:
                n_in += symbols_in.count(species)
                n_ref += symbols_ref.count(species)
            atvol_in = input_structure.get_volume() / n_in
            atvol_ref = reference_structure.get_volume() / n_ref
            scale = atvol_in / atvol_ref
    modcell *= (1.0 / scale) ** (1.0 / 3.0)
    return modcell


def _get_transformation_matrix(input_cell: np.ndarray,
                               reference_cell: np.ndarray,
                               tolerance_cell: float = 0.05) -> np.ndarray:
    """
    Obtain the(in general non - integer) transformation matrix connecting the
    input structure to the reference structure L=L_p.P - -> P=L_p ^ -1.L

    Parameters
    ----------
    input_cell
        Cell metric of input structure(possibly scaled)
    reference_cell
        Cell metric of reference structure
    tolerance_cell
        Tolerance for how much the elements of P are allowed to deviate from
        nearest integer before they are rounded.

    Returns
    -------
    Transformation matrix P of integers.
    """
    P = np.dot(input_cell, np.linalg.inv(reference_cell))

    # assert that the transformation matrix does not deviate too
    # strongly from the nearest integer matrix
    assert np.linalg.norm(P - np.around(P)) / 9 < tolerance_cell, \
        ('Failed to map structure to reference' +
         'structure (tolerance_cell exceeded).\n' +
         'reference:\n {}\n'.format(reference_cell) +
         'input:\n {}\n'.format(input_cell) +
         'P:\n {}\n'.format(P) +
         'det P = {}\n'.format(np.linalg.det(P)) +
         'P_round:\n {}\n'.format(np.around(P)) +
         'Deviation: {}\n'.format(np.linalg.norm(P - np.around(P)) / 9) +
         'If there are vacancies, you can try specifying `inert_species`.' +
         ' Else, you can try raising `tolerance_cell`.')

    # reduce the (real) transformation matrix to the nearest integer one
    P = np.around(P)
    return P


def _rescale_structures(input_structure: Atoms,
                        reference_structure: Atoms,
                        P: np.ndarray,
                        tolerance_positions: float = 0.01) \
                        -> Tuple[Atoms, Atoms]:
    """
    Rescale `input_structure` with `P` so that it matches
    `reference_structure`, and make a supercell of `reference_structure` using
    `P`

    Parameters
    ----------
    input_structure
        relaxed input structure
    reference_structure
        reference structure, which can but need not represent the primitive
        cell
    P
        Transformation matrix of integers
    tolerance_positions
        tolerance factor applied when scanning for overlapping positions in
        Angstrom(forwarded to `ase.build.cut`)

    Returns
    -------
    A tuple with the scaled version of `input_structure` and the supercell of
    `reference_structure` matching cell metric of `scaled_structure`.
    """
    scaled_structure = input_structure.copy()
    scaled_structure.set_cell(np.dot(P, reference_structure.cell),
                              scale_atoms=True)

    # generate supercell of (presumably primitive) reference structure
    ideal_supercell = cut(reference_structure,
                          P[0], P[1], P[2],
                          tolerance=tolerance_positions)
    assert(len(ideal_supercell) ==
           int(np.round(len(reference_structure) * np.linalg.det(P)))), \
        ('Supercell construction of reference structure failed'
         ' (number of atoms do not match).\n'
         'Permutation matrix used:\n{}'.format(P) +
         '\nYou can try change tolerance_positions.')

    return scaled_structure, ideal_supercell



# -*- coding: utf-8 -*-

from .canonical_ensemble import CanonicalEnsemble
from .semi_grand_canonical_ensemble import SemiGrandCanonicalEnsemble
from .vcsgc_ensemble import VCSGCEnsemble

__all__ = ['CanonicalEnsemble',
           'SemiGrandCanonicalEnsemble',
           'VCSGCEnsemble']



import getpass
import json
import numbers
import numpy as np
import pandas as pd
import tarfile
import tempfile
import socket

from ase import Atoms
from ase.io import Trajectory
from ase.io import write as ase_write, read as ase_read
from collections import OrderedDict
from datetime import datetime
from typing import BinaryIO, Dict, List, TextIO, Tuple, Union
from icet import __version__ as icet_version


class DataContainer:
    """
    Data container for storing information concerned with
    Monte Carlo simulations performed with mchammer.

    Parameters
    ----------
    atoms : ASE Atoms object
        reference atomic structure associated with the data container

    ensemble_parameters : dict
        parameters associated with the underlying ensemble

    metadata : dict
        metadata associated with the data container
    """

    def __init__(self, atoms: Atoms, ensemble_parameters: dict,
                 metadata: dict = OrderedDict()):
        """
        Initializes a DataContainer object.
        """
        if not isinstance(atoms, Atoms):
            raise TypeError('atoms is not an ASE Atoms object')

        self.atoms = atoms.copy()
        self._ensemble_parameters = ensemble_parameters
        self._metadata = metadata
        self._add_default_metadata()
        self._last_state = {}

        self._data = pd.DataFrame(columns=['mctrial'])

    def append(self, mctrial: int,
               record: Dict[str, Union[int, float, list]]):
        """
        Appends data to data container.

        Parameters
        ----------
        mctrial
            current Monte Carlo trial step
        record
            dictionary of tag-value pairs representing observations

        Raises
        ------
        TypeError
            if input parameters have the wrong type

        Todo
        ----
        * This might be a quite expensive way to add data to the data
          frame. Testing and profiling to be carried out later.
        """
        if not isinstance(mctrial, numbers.Integral):
            raise TypeError('mctrial has the wrong type: {}'
                            .format(type(mctrial)))
        if not self._data.mctrial.empty:
            if self._data.mctrial.iloc[-1] > mctrial:
                raise ValueError('mctrial values should be given in ascending'
                                 ' order. This error can for example occur'
                                 ' when trying to append to an existing data'
                                 ' container after having reset the time step.'
                                 ' Note that the latter happens automatically'
                                 ' when initializing a new ensemble.')

        if not isinstance(record, dict):
            raise TypeError('record has the wrong type: {}'
                            .format(type(record)))
        row_data = OrderedDict()
        row_data['mctrial'] = mctrial
        row_data.update(record)
        self._data = self._data.append(row_data, ignore_index=True)
        self._data.mctrial = self._data.mctrial.astype('int')

    def _update_last_state(self, last_step: int, occupations: List[int],
                           accepted_trials: int, random_state: tuple):
        """Updates last state of the simulation: last step, occupation vector
        and number of accepted trial steps.

        Parameters
        ----------
        last_step
            last trial step
        occupations
            occupation vector observed during the last trial step
        accepted_trial
            number of current accepted trial steps
        random_state
            tuple representing the last state of the random generator
        """
        self._last_state['last_step'] = last_step
        self._last_state['occupations'] = occupations
        self._last_state['accepted_trials'] = accepted_trials
        self._last_state['random_state'] = random_state

    def get_data(self, *tags,
                 start: int = None,
                 stop: int = None,
                 interval: int = 1,
                 fill_method: str = 'skip_none',
                 apply_to: List[str] = None) \
            -> Union[np.ndarray, List[Atoms], Tuple[np.ndarray, List[Atoms]]]:
        """Returns the accumulated data for the requested observables.

        Parameters
        ----------
        tags
            tuples of the requested properties

        start
            minimum value of trial step to consider; by default the
            smallest value in the mctrial column will be used.

        stop
            maximum value of trial step to consider; by default the
            largest value in the mctrial column will be used.

        interval
            increment for mctrial; by default the smallest available
            interval will be used.

        fill_method : {'skip_none', 'fill_backward', 'fill_forward',
                       'linear_interpolate', None}
            method employed for dealing with missing values; by default
            uses 'skip_none'.

        apply_to
            tags of columns for which fill_method will be employed;
            by default parse all columns with fill_method.

        Raises
        ------
        ValueError
            if tags is empty
        ValueError
            if observables are requested that are not in data container
        ValueError
            if fill method is unknown
        ValueError
            if trajectory is requested and fill method is not skip_none
        """
        fill_methods = ['skip_none',
                        'fill_backward',
                        'fill_forward',
                        'linear_interpolate']

        if len(tags) == 0:
            raise TypeError('Missing tags argument')

        if 'trajectory' in tags:
            if fill_method != 'skip_none':
                raise ValueError('Only skip_none fill method is avaliable'
                                 ' when trajectory is requested')
            return self._get_trajectory(*tags, start=start, stop=stop,
                                        interval=interval)

        for tag in tags:
            if tag not in self._data:
                raise ValueError('No observable named {} in data'
                                 ' container'.format(tag))

        if start is None and stop is None:
            data = self._data.loc[::interval, tags]
        else:
            data = self._data.set_index(self._data.mctrial)
            # slice and pass a copy to avoid slowing down dropna method below
            if start is None:
                data = data.loc[:stop:interval, tags].copy()
            elif stop is None:
                data = data.loc[start::interval, tags].copy()
            else:
                data = data.loc[start:stop:interval, tags].copy()

        if fill_method is not None:
            if fill_method not in fill_methods:
                raise ValueError('Unknown fill method: {}'
                                 .format(fill_method))

            if apply_to is None:
                apply_to = tags

            # retrieve only valid observations
            if fill_method is 'skip_none':
                data.dropna(inplace=True, subset=apply_to)

            else:
                # if requested, drop NaN values in columns
                subset = [tag for tag in tags if tag not in apply_to]
                data.dropna(inplace=True, subset=subset)

                # fill NaN with the next valid observation
                if fill_method is 'fill_backward':
                    data.fillna(method='bfill', inplace=True)

                # fill NaN with the last valid observation
                elif fill_method is 'fill_forward':
                    data.fillna(method='ffill', inplace=True)

                # fill NaN with the linear interpolation of the last and
                # next valid observations
                elif fill_method is 'linear_interpolate':
                    data.interpolate(limit_area='inside', inplace=True)

                # drop any left-over nan values
                data.dropna(inplace=True)

        data_list = []
        for tag in tags:
            # convert NaN to None
            data_list.append(np.array(
                [None if np.isnan(x).any() else x for x in data[tag]]))
        if len(tags) > 1:
            # return a tuple if more than one tag is given
            return tuple(data_list)
        else:
            # return a list if only one tag is given
            return data_list[0]

    @property
    def data(self) -> pd.DataFrame:
        """ pandas data frame (see :class:`pandas.DataFrame`) """
        return self._data

    @property
    def ensemble_parameters(self) -> dict:
        """ parameters associated with Monte Carlo simulation """
        return self._ensemble_parameters.copy()

    @property
    def observables(self) -> List[str]:
        """ observable names """
        return [col for col in self._data.columns.tolist() if col != 'mctrial']

    @property
    def metadata(self) -> dict:
        """ metadata associated with data container """
        return self._metadata

    @property
    def last_state(self) -> Dict[str, Union[int, List[int]]]:
        """ last state to be used to restart Monte Carlo simulation """
        return self._last_state

    def reset(self):
        """ Resets (clears) data frame of data container. """
        self._data = pd.DataFrame(columns=['mctrial'])
        self._data = self._data.astype({'mctrial': int})

    def get_number_of_entries(self, tag: str = None) -> int:
        """
        Returns the total number of entries with the given observable tag.

        Parameters
        ----------
        tag
            name of observable; by default the total number of rows in the
            data frame will be returned.

        Raises
        ------
        ValueError
            if observable is requested that is not in data container
        """
        if tag is None:
            return len(self._data)
        else:
            if tag not in self._data:
                raise ValueError('No observable named {}'
                                 ' in data container'.format(tag))
            return self._data[tag].count()

    def get_average(self, tag: str,
                    start: int = None, stop: int = None) -> float:
        """
        Returns average of a scalar observable.

        Parameters
        ----------
        tag
            tag of field over which to average
        start
            minimum value of trial step to consider; by default the
            smallest value in the mctrial column will be used.
        stop
            maximum value of trial step to consider; by default the
            largest value in the mctrial column will be used.

        Raises
        ------
        ValueError
            if observable is requested that is not in data container
        ValueError
            if observable is not scalar
        """
        if tag in ['trajectory', 'occupations']:
            raise ValueError('{} is not scalar'.format(tag))
        data = self.get_data(tag, start=start, stop=stop)
        return np.mean(data)

    def get_standard_deviation(self, tag: str, start: int = None,
                               stop: int = None) -> float:
        """
        Returns standard deviation of a scalar observable, calculated using
        numpy.

        Parameters
        ----------
        tag
            tag of field over which to average
        start
            minimum value of trial step to consider; by default the
            smallest value in the mctrial column will be used.
        stop
            maximum value of trial step to consider; by default the
            largest value in the mctrial column will be used.

        Raises
        ------
        ValueError
            if observable is requested that is not in data container
        ValueError
            if observable is not scalar
        """
        if tag in ['trajectory', 'occupations']:
            raise ValueError('{} is not scalar'.format(tag))
        data = self.get_data(tag, start=start, stop=stop)
        return np.std(data)

    def _get_trajectory(self, *tags, start: int = None, stop: int = None,
                        interval: int = 1) \
            -> Union[List[Atoms], Tuple[List[Atoms], np.ndarray]]:
        """
        Returns a trajectory in the form of a list of ASE Atoms
        along with the corresponding values of the mctrial and/or scalar
        properties upon request.
        Configurations with non properties will be
        skipped in the trajectory if the property is requested.

        Parameters
        ----------
        start
            minimum value of trial step to consider; by default the
            smallest value in the mctrial column will be used.
        stop
            maximum value of trial step to consider; by default the
            largest value in the mctrial column will be used.
        interval
            increment for mctrial; by default the smallest available
            interval will be used.
        """
        new_tags = tuple(['occupations' if tag == 'trajectory' else
                          tag for tag in tags])
        data = \
            self.get_data(*new_tags, start=start, stop=stop, interval=interval)

        if len(tags) > 1:
            data_list = list(data)
        else:
            data_list = [data]

        tag_list = list(new_tags)
        atoms_list = []
        for tag, data_row in zip(tag_list, data_list):
            if tag == 'occupations':
                ind = tag_list.index('occupations')
                for occupation_vector in data_row:
                    atoms = self.atoms.copy()
                    atoms.numbers = occupation_vector
                    atoms_list.append(atoms)
                data_list[ind] = atoms_list

        if len(data_list) > 1:
            return tuple(data_list)
        else:
            return data_list[0]

    def write_trajectory(self, outfile: Union[str, BinaryIO, TextIO]):
        """
        Saves trajectory to a file along with the respectives values of the
        potential field for each configuration. If the file exists the
        trajectory will be appended. Use ase gui to visualize the trajectory
        with values of the potential for each frame.

        Parameters
        ----------
        outfile
            output file name or file object
        """
        atoms_list, energies = self._get_trajectory('occupations', 'potential')
        traj = Trajectory(outfile, mode='a')
        for atoms, energy in zip(atoms_list, energies):
            traj.write(atoms=atoms, energy=energy)
        traj.close()

    @staticmethod
    def read(infile: Union[str, BinaryIO, TextIO]):
        """
        Reads DataContainer object from file.

        Parameters
        ----------
        infile
            file from which to read

        Raises
        ------
        FileNotFoundError
            if file is not found (str)
        ValueError
            if file is of incorrect type (not a tarball)
        """
        if isinstance(infile, str):
            filename = infile
        else:
            filename = infile.name

        if not tarfile.is_tarfile(filename):
            raise TypeError('{} is not a tar file'.format(filename))

        reference_atoms_file = tempfile.NamedTemporaryFile()
        reference_data_file = tempfile.NamedTemporaryFile()
        runtime_data_file = tempfile.NamedTemporaryFile()

        with tarfile.open(mode='r', name=filename) as tar_file:
            # file with atoms
            reference_atoms_file.write(tar_file.extractfile('atoms').read())

            reference_atoms_file.seek(0)
            atoms = ase_read(reference_atoms_file.name, format='json')

            # file with reference data
            reference_data_file.write(
                tar_file.extractfile('reference_data').read())
            reference_data_file.seek(0)
            with open(reference_data_file.name, encoding='utf-8') as fd:
                reference_data = json.load(fd)

            # init DataContainer
            dc = \
                DataContainer(atoms=atoms,
                              ensemble_parameters=reference_data['parameters'])

            # overwrite metadata
            dc._metadata = reference_data['metadata']

            for tag, value in reference_data['last_state'].items():
                if tag == 'random_state':
                    value = \
                        tuple(tuple(x) if isinstance(x, list)
                              else x for x in value)
                dc._last_state[tag] = value

            # add runtime data from file
            runtime_data_file.write(
                tar_file.extractfile('runtime_data').read())

            runtime_data_file.seek(0)
            runtime_data = pd.read_json(runtime_data_file)
            dc._data = runtime_data.sort_index(ascending=True)
            dc._data.occupations = \
                dc._data.occupations.replace({None: np.nan})

        return dc

    def _write(self, outfile: Union[str, BinaryIO, TextIO]):
        """
        Writes DataContainer object to file.

        Parameters
        ----------
        outfile
            file to which to write
        """
        self._metadata['date_last_backup'] = \
            datetime.now().strftime('%Y-%m-%dT%H:%M:%S')

        # Save reference atomic structure
        reference_atoms_file = tempfile.NamedTemporaryFile()
        ase_write(reference_atoms_file.name, self.atoms, format='json')

        # Save reference data
        reference_data = {'parameters': self._ensemble_parameters,
                          'metadata': self._metadata,
                          'last_state': self._last_state}

        reference_data_file = tempfile.NamedTemporaryFile()
        with open(reference_data_file.name, 'w') as handle:
            json.dump(reference_data, handle)

        # Save pandas DataFrame
        runtime_data_file = tempfile.NamedTemporaryFile()
        self._data.to_json(runtime_data_file.name, double_precision=15)

        with tarfile.open(outfile, mode='w') as handle:
            handle.add(reference_atoms_file.name, arcname='atoms')
            handle.add(reference_data_file.name, arcname='reference_data')
            handle.add(runtime_data_file.name, arcname='runtime_data')
        runtime_data_file.close()

    def _add_default_metadata(self):
        """Adds default metadata to metadata dict."""

        self._metadata['date_created'] = \
            datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
        self._metadata['username'] = getpass.getuser()
        self._metadata['hostname'] = socket.gethostname()
        self._metadata['icet_version'] = icet_version



import tarfile
import tempfile

from typing import BinaryIO, Dict, List, TextIO, Tuple, Union

import numpy as np
import ase.db
from ase import Atoms

from icet import ClusterSpace
from icet.io.logging import logger
logger = logger.getChild('structure_container')


class StructureContainer:
    """
    This class serves as a container for structure objects as well as their fit
    properties and cluster vectors.

    Parameters
    ----------
    cluster_space : icet.ClusterSpace
        cluster space used for evaluating the cluster vectors

    list_of_atoms : list or tuple or list(tuple)
        list of atoms; if the list contains tuples, the second element of the
        tuple will be used as a tag of the structure

    list_of_properties : list(dict)
        list of properties, which are provided in dicts
    """

    def __init__(self, cluster_space: ClusterSpace,
                 list_of_atoms: Union[list, tuple] = None,
                 list_of_properties: List[dict] = None,
                 allow_duplicate: bool = True):

        self._cluster_space = cluster_space
        self._structure_list = []

        # Add atomic structures
        if list_of_atoms is not None:
            if not isinstance(list_of_atoms, (list, tuple)):
                raise TypeError('atoms must be given as a list or a tuple')

            if list_of_properties is not None:
                if not len(list_of_properties) == len(list_of_atoms):
                    raise ValueError('list of atoms and list of properties'
                                     ' must have the same length')
            else:
                list_of_properties = [None] * len(list_of_atoms)

            if not all(isinstance(x, tuple) for x in list_of_atoms):
                list_of_atoms = [(atoms, None) for atoms in list_of_atoms]

            for (atoms, user_tag), properties in zip(list_of_atoms,
                                                     list_of_properties):
                try:
                    self.add_structure(atoms=atoms, user_tag=user_tag,
                                       properties=properties,
                                       allow_duplicate=allow_duplicate)
                except (TypeError, ValueError):
                    logger.exception('Skipping structure {}; '.format(
                        list_of_atoms.index((atoms, user_tag))))

    def __len__(self) -> int:
        return len(self._structure_list)

    def __getitem__(self, ind: int):
        return self._structure_list[ind]

    def get_structure_indices(self, user_tag: str = None) -> List[int]:
        """
        Get structure indices via user_tag

        Parameters
        ----------
        user_tag
            user_tag used for selecting structures

        Returns
        -------
        list of integers
            List of structure's indices
        """
        return [i for i, s in enumerate(self)
                if user_tag is None or s.user_tag == user_tag]

    def _get_string_representation(self, print_threshold: int = None,
                                   print_minimum: int = 10) -> str:
        """
        String representation of the structure container that provides an
        overview of the structures in the container.

        Parameters
        ----------
        print_threshold
            if the number of structures exceeds this number print dots
        print_minimum
            number of lines printed from the top and the bottom of the
            structure list if `print_threshold` is exceeded

        Returns
        -------
        multi-line string
            string representation of the structure container
        """

        def repr_structure(structure, index=-1, header=False):
            from collections import OrderedDict
            fields = OrderedDict([
                ('index',     '{:4}'.format(index)),
                ('user_tag',  '{:21}'.format(structure.user_tag)),
                ('natoms',    '{:5}'.format(len(structure))),
                ('chemical formula', structure._atoms.get_chemical_formula())])
            fields.update(sorted(structure.properties.items()))
            for key, value in fields.items():
                if isinstance(value, float):
                    fields[key] = '{:8.3f}'.format(value)
                if isinstance(value, int):
                    fields[key] = '{:8}'.format(value)
            s = []
            for name, value in fields.items():
                n = max(len(name), len(value))
                if header:
                    s += ['{s:^{n}}'.format(s=name, n=n)]
                else:
                    if name == 'user_tag' or name == 'chemical formula':
                        # We want them aligned to the left
                        value = '{:{padding}}'.format(value, padding=n - 1)
                    s += ['{s:^{n}}'.format(s=value, n=n)]
            return ' | '.join(s)

        if len(self) == 0:
            return 'Empty StructureContainer'

        # basic information
        # (use last structure in list to obtain maximum line length)
        dummy = self._structure_list[-1]
        width = len(repr_structure(dummy))

        # table header
        s = []  # type: List
        s += ['{s:=^{n}}'.format(s=' Structure Container ', n=width)]
        s += ['Total number of structures: {}'.format(len(self))]
        s += [''.center(width, '-')]
        s += [repr_structure(dummy, header=True).rstrip()]
        s += [''.center(width, '-')]

        # table body
        index = 0
        while index < len(self):
            if (print_threshold is not None and
                    len(self) > print_threshold and
                    index >= print_minimum and
                    index <= len(self) - print_minimum):
                index = len(self) - print_minimum
                s += [' ...']
            s += [repr_structure(self._structure_list[index], index=index)]
            index += 1
        s += [''.center(width, '=')]

        return '\n'.join(s)

    def __repr__(self) -> str:
        """ String representation. """
        return self._get_string_representation(print_threshold=50)

    def print_overview(self, print_threshold: int = None,
                       print_minimum: int = 10):
        """
        Prints a list of structures in the structure container.

        Parameters
        ----------
        print_threshold
            if the number of orbits exceeds this number print dots
        print_minimum
            number of lines printed from the top and the bottom of the orbit
            list if `print_threshold` is exceeded
        """
        print(self._get_string_representation(print_threshold=print_threshold,
                                              print_minimum=print_minimum))

    def add_structure(self, atoms: Atoms, user_tag: str = None,
                      properties: dict = None,
                      allow_duplicate: bool = True):
        """
        Adds a structure to the structure container.

        Parameters
        ----------
        atoms
            the atomic structure to be added
        user_tag
            custom user tag to label structure
        properties
            scalar properties. If properties are not specified the atoms
            object are required to have an attached ASE calculator object
            with a calculated potential energy
        allow_duplicate
             whether or not to add the structure if there already exists a
             structure with identical cluster-vector
        """
        # atoms must have a proper format and label
        if not isinstance(atoms, Atoms):
            raise TypeError('atoms must be an ASE Atoms object.'
                            ' Not {}'.format(type(atoms)))

        if user_tag is not None:
            if not isinstance(user_tag, str):
                raise TypeError('user_tag must be a string.'
                                ' Not {}.'.format(type(user_tag)))

        atoms_copy = atoms.copy()

        # check for properties in attached calculator
        if properties is None and atoms.calc:
            properties = {}
            if not atoms.calc.calculation_required(atoms, ['energy']):
                energy = atoms.get_potential_energy()
                properties['energy'] = energy / len(atoms)

        # check if there exists structures with identical cluster vector
        cv = self._cluster_space.get_cluster_vector(atoms_copy)
        if not allow_duplicate:
            for i, fs in enumerate(self):
                if np.allclose(cv, fs.cluster_vector):
                    msg = "{} have identical cluster vector with {}".format(
                        user_tag if user_tag is not None else 'Input atoms',
                        fs.user_tag if fs.user_tag != 'None' else 'structure')
                    msg += " at index {}".format(i)
                    raise ValueError(msg)

        structure = FitStructure(atoms_copy, user_tag)
        structure.set_properties(properties)
        structure._set_cluster_vector(cv)
        self._structure_list.append(structure)

    def get_condition_number(self, structure_indices: List[int] = None,
                             key: str = 'energy') -> float:
        """ Returns the condition number for the sensing matrix.

        A very large condition number can be a sign of multicollinearity,
        read more here https://en.wikipedia.org/wiki/Condition_number

        Parameters
        ----------
        structure_indices
            list of structure indices. By default (``None``) the
            method will return all fit data available.
        key
            key of properties dictionary

        Returns
        -------
        condition number for the sensing matrix
        """
        return np.linalg.cond(self.get_fit_data(structure_indices, key)[0])

    def get_fit_data(self, structure_indices: List[int] = None,
                     key: str = 'energy') -> Tuple[np.ndarray, np.ndarray]:
        """
        Returns fit data for all structures. The cluster vectors and
        target properties for all structures are stacked into NumPy arrays.

        Parameters
        ----------
        structure_indices
            list of structure indices. By default (``None``) the
            method will return all fit data available.
        key
            key of properties dictionary

        Returns
        -------
        cluster vectors and target properties for desired structures
        """
        if structure_indices is None:
            cv_list = [s.cluster_vector
                       for s in self._structure_list]
            prop_list = [s.properties[key]
                         for s in self._structure_list]
        else:
            cv_list, prop_list = [], []
            for i in structure_indices:
                cv_list.append(self._structure_list[i].cluster_vector)
                prop_list.append(self._structure_list[i].properties[key])

        if len(cv_list) == 0:
            raise Exception('No available fit data for'
                            ' {}'.format(structure_indices))

        return np.array(cv_list), np.array(prop_list)

    def add_properties(self, structure_indices: List[int] = None,
                       properties: List[dict] = None):
        """
        This method allows you to add properties and/or modify
        the values of existing properties

        Parameters
        ----------
        structure_indices
            list of structure indices. By default (``None``) the
            method will add the properties to all structures.

        properties
            list of scalar properties
        """
        if structure_indices is None:
            msg = 'len of properties does not equal len of fit structures'
            assert len(properties) == len(self), msg
            for s, prop in zip(self._structure_list, properties):
                s.set_properties(prop)
        else:
            for i, prop in zip(structure_indices, properties):
                self._structure_list[i].set_properties(prop)

    def get_properties(self, structure_indices: List[int] = None,
                       key: str = 'energy') -> List[float]:
        """
        Returns a list with the value of properties with key='key'
        for a desired set of structures

        Parameters
        ----------
        structures_indices
            list of structure indices. Default to
            None and in that case returns properties of all structures

        key
            key of properties dictionary. Default to 'energy'
        """
        if structure_indices is None:
            prop_list = [s.properties[key]
                         for s in self._structure_list]
        else:
            prop_list = []
            for i in structure_indices:
                prop_list.append(self._structure_list[i].properties[key])

        return prop_list

    def get_structures(self,
                       structure_indices: List[int] = None) -> List[Atoms]:
        """
        Returns a list of structures in the form of ASE Atoms

        Parameters
        ----------
        structure_indices
            list of structure indices. By default (``None``) the
            method will return all structures listed in the container
        """
        if structure_indices is None:
            s_list = [s.atoms for s in self._structure_list]
        else:
            s_list = []
            for i in structure_indices:
                s_list.append(self._structure_list[i].atoms)

        return s_list

    def get_user_tags(self, structure_indices: List[int] = None) -> List[str]:
        """
        Return a list of user tags for the structures in the structure
        container

        Parameters
        ----------
        structure_indices
            list of structure indices. By default (``None``) the
            method will return all user tags listed in the container
        """
        if structure_indices is None:
            tag_list = [s.user_tag for s in self._structure_list]
        else:
            tag_list = []
            for i in structure_indices:
                tag_list.append(self._structure_list[i].user_tag)

        return tag_list

    @property
    def cluster_space(self) -> ClusterSpace:
        """Cluster space used to calculate the cluster vectors."""
        return self._cluster_space

    @property
    def fit_structures(self):
        """List of the fit structures."""
        return self._structure_list

    @property
    def available_properties(self) -> List[str]:
        """List of the available properties."""
        return sorted(set([p for fs in self for p in fs.properties.keys()]))

    def write(self, outfile: Union[str, BinaryIO, TextIO]):
        """
        Writes structure container to a file.

        Parameters
        ----------
        outfile
            output file name or file object
        """
        # Write cluster space to tempfile
        temp_cs_file = tempfile.NamedTemporaryFile()
        self.cluster_space.write(temp_cs_file.name)

        # Write fit structures as an ASE db in tempfile
        temp_db_file = tempfile.NamedTemporaryFile()
        db = ase.db.connect(temp_db_file.name, type='db', append=False)

        for fit_structure in self.fit_structures:
            data_dict = {'user_tag': fit_structure.user_tag,
                         'properties': fit_structure.properties,
                         'cluster_vector': fit_structure.cluster_vector}
            db.write(fit_structure.atoms, data=data_dict)

        with tarfile.open(outfile, mode='w') as handle:
            handle.add(temp_db_file.name, arcname='database')
            handle.add(temp_cs_file.name, arcname='cluster_space')

    @staticmethod
    def read(infile: Union[str, BinaryIO, TextIO]):
        """
        Reads StructureContainer object from file.

        Parameters
        ----------
        infile
            file from which to read

        """
        if isinstance(infile, str):
            filename = infile
        else:
            filename = infile.name

        if not tarfile.is_tarfile(filename):
            raise TypeError('{} is not a tar file'.format(filename))

        temp_db_file = tempfile.NamedTemporaryFile()
        with tarfile.open(mode='r', name=filename) as tar_file:
            cs_file = tar_file.extractfile('cluster_space')
            temp_db_file.write(tar_file.extractfile('database').read())
            temp_db_file.seek(0)
            cluster_space = ClusterSpace.read(cs_file)
            database = ase.db.connect(temp_db_file.name, type='db')

            structure_container = StructureContainer(cluster_space)
            fit_structures = []
            for row in database.select():
                data = row.data
                fit_structure = FitStructure(row.toatoms(),
                                             user_tag=data['user_tag'],
                                             cv=data['cluster_vector'],
                                             properties=data['properties'])
                fit_structures.append(fit_structure)
            structure_container._structure_list = fit_structures

        return structure_container


class FitStructure:
    """
    This class holds a supercell along with its properties and cluster
    vector.

    Attributes
    ----------
    atoms : ASE Atoms
        supercell structure
    user_tag : str
        custom user tag
    cvs : NumPy array
        calculated cluster vector for actual structure
    properties : dict
        the properties dictionary
    """

    def __init__(self, atoms: Atoms, user_tag: str,
                 cv: np.ndarray = None, properties: dict = None):
        self._atoms = atoms
        self._user_tag = user_tag
        self._properties = {}
        self._set_cluster_vector(cv)
        self.set_properties(properties)

    @property
    def cluster_vector(self) -> np.ndarray:
        """calculated cluster vector"""
        return self._cluster_vector

    @property
    def atoms(self) -> Atoms:
        """supercell structure"""
        return self._atoms

    @property
    def user_tag(self) -> str:
        """structure label"""
        return str(self._user_tag)

    @property
    def properties(self) -> Dict[str, Union[float, int]]:
        """scalar properties of structure"""
        return self._properties

    def __getattr__(self, key):
        """Accesses properties if possible and returns value"""
        if key not in self.properties.keys():
            return super().__getattribute__(key)
        return self.properties[key]

    def _set_cluster_vector(self, cv: np.ndarray):
        """ Sets the cluster vectors of the structure. """
        if cv is not None:
            self._cluster_vector = cv
        else:
            self._cluster_vector = None

    def set_properties(self, properties: Dict[str, float]):
        """ Sets properties associated with the structure. """
        if properties is not None:
            self._properties.update(properties)

    def __len__(self) -> int:
        """ Number of sites in the structure. """
        return len(self._atoms)



import pickle
from collections import OrderedDict
from typing import List, Union

import numpy as np

from _icet import ClusterSpace as _ClusterSpace
from ase import Atoms
from icet.core.orbit_list import OrbitList
from icet.core.structure import Structure
from icet.tools.geometry import (add_vacuum_in_non_pbc,
                                 get_decorated_primitive_structure)


class ClusterSpace(_ClusterSpace):
    """
    This class provides functionality for generating and maintaining cluster
    spaces.

    Parameters
    ----------
    atoms : ase.Atoms
        atomic configuration
    cutoffs : list(float)
        cutoff radii per order that define the cluster space
    chemical_symbols : list(str) or list(list(str))
        list of chemical symbols, each of which must map to an element
        of the periodic table

        If a list of chemical symbols is provided, all sites on the
        lattice will have the same allowed occupations as the input
        list.

        If a list of list of chemical symbols is provided then the
        outer list must be the same length as the atoms object and
        ``chemical_symbols[i]`` will correspond to the allowed species
        on lattice site ``i``.
    """

    def __init__(self, atoms: Atoms, cutoffs: List[float],
                 chemical_symbols: Union[List[str], List[List[str]]]) -> None:

        if not isinstance(atoms, Atoms):
            raise TypeError('Input configuration must be an ASE Atoms object'
                            ', not type {}'.format(type(atoms)))

        self._atoms = atoms.copy()
        self._cutoffs = cutoffs
        self._chemical_symbols = chemical_symbols.copy()
        self._input_chemical_symbols = chemical_symbols.copy()

        # handle occupations
        if all(isinstance(i, str) for i in self._chemical_symbols):
            self._chemical_symbols = [self._chemical_symbols]*len(self._atoms)

        elif not all(isinstance(i, list) for i in self._chemical_symbols):
            raise TypeError("chemical_symbols must be"
                            " List[str] or List[List[str]],"
                            " not {}".format(type(self._chemical_symbols)))

        decorated_primitive, primitive_chemical_symbols = \
            get_decorated_primitive_structure(
                self._atoms, self._chemical_symbols)

        self._primitive_chemical_symbols = primitive_chemical_symbols

        assert len(decorated_primitive) == len(primitive_chemical_symbols)

        # set up orbit list
        self._orbit_list = OrbitList(decorated_primitive, self._cutoffs)
        self._orbit_list.remove_inactive_orbits(primitive_chemical_symbols)

        # call (base) C++ constructor
        _ClusterSpace.__init__(
            self, primitive_chemical_symbols, self._orbit_list)

    def _get_chemical_symbol_representation(self):
        """Returns a str version of the chemical symbols that is
        easier on the eyes.
        """
        nice_str = ''
        if len(self.chemical_symbols) > 4:
            last_symbol = self.chemical_symbols[0]
            count = 1
            for i in range(1, len(self.chemical_symbols)):
                if self.chemical_symbols[i] == last_symbol:
                    count += 1
                    if i == len(self.chemical_symbols)-1:
                        if count == 1:
                            nice_str += '{} '.format(last_symbol)
                        else:
                            nice_str += '{}*{} '.format(count, last_symbol)
                else:
                    if count == 1:
                        nice_str += '{} '.format(last_symbol)
                    else:
                        nice_str += '{}*{} '.format(count, last_symbol)
                    count = 1
                    last_symbol = self.chemical_symbols[i]
        else:
            for s in self.chemical_symbols:
                nice_str += '{} '.format(s)
        return nice_str

    def _get_string_representation(self, print_threshold: int = None,
                                   print_minimum: int = 10) -> str:
        """
        String representation of the cluster space that provides an overview of
        the orbits (order, radius, multiplicity etc) that constitute the space.

        Parameters
        ----------
        print_threshold
            if the number of orbits exceeds this number print dots
        print_minimum
            number of lines printed from the top and the bottom of the orbit
            list if `print_threshold` is exceeded

        Returns
        -------
        multi-line string
            string representation of the cluster space.
        """

        def repr_orbit(orbit, header=False):
            formats = {'order': '{:2}',
                       'radius': '{:8.4f}',
                       'multiplicity': '{:4}',
                       'index': '{:4}',
                       'orbit_index': '{:4}',
                       'multi_component_vector': '{:}'}
            s = []
            for name, value in orbit.items():
                str_repr = formats[name].format(value)
                n = max(len(name), len(str_repr))
                if header:
                    s += ['{s:^{n}}'.format(s=name, n=n)]
                else:
                    s += ['{s:^{n}}'.format(s=str_repr, n=n)]
            return ' | '.join(s)

        # basic information
        # (use largest orbit to obtain maximum line length)
        prototype_orbit = self.orbit_data[-1]
        width = len(repr_orbit(prototype_orbit))
        s = []  # type: List
        s += ['{s:=^{n}}'.format(s=' Cluster Space ', n=width)]
        s += [' chemical species: {}'
              .format(self._get_chemical_symbol_representation())]
        s += [' cutoffs: {}'.format(' '.join(['{:.4f}'.format(co)
                                              for co in self._cutoffs]))]
        s += [' total number of orbits: {}'.format(len(self))]
        t = ['{}= {}'.format(k, c)
             for k, c in self.get_number_of_orbits_by_order().items()]
        s += [' number of orbits by order: {}'.format('  '.join(t))]

        # table header
        s += [''.center(width, '-')]
        s += [repr_orbit(prototype_orbit, header=True)]
        s += [''.center(width, '-')]

        # table body
        index = 0
        orbit_list_info = self.orbit_data
        while index < len(orbit_list_info):
            if (print_threshold is not None and
                    len(self) > print_threshold and
                    index >= print_minimum and
                    index <= len(self) - print_minimum):
                index = len(self) - print_minimum
                s += [' ...']
            s += [repr_orbit(orbit_list_info[index])]
            index += 1
        s += [''.center(width, '=')]

        return '\n'.join(s)

    def __repr__(self) -> str:
        """ String representation. """
        return self._get_string_representation(print_threshold=50)

    def print_overview(self,
                       print_threshold: int = None, print_minimum: int = 10):
        """
        Print an overview of the cluster space in terms of the orbits (order,
        radius, multiplicity etc).

        Parameters
        ----------
        print_threshold
            if the number of orbits exceeds this number print dots
        print_minimum
            number of lines printed from the top and the bottom of the orbit
            list if `print_threshold` is exceeded
        """
        print(self._get_string_representation(print_threshold=print_threshold,
                                              print_minimum=print_minimum))

    @property
    def orbit_data(self):
        """
        list of dicts : list of orbits ith information regarding
        order, radius, multiplicity etc
        """
        data = []
        zerolet = OrderedDict([('index', 0),
                               ('order', 0),
                               ('radius', 0),
                               ('multiplicity', 1),
                               ('orbit_index', -1),
                               ('multi_component_vector', '.')])

        data.append(zerolet)
        index = 1
        while index < len(self):
            cluster_space_info = self.get_cluster_space_info(index)
            orbit_index = cluster_space_info[0]
            mc_vector = cluster_space_info[1]
            orbit = self.get_orbit(orbit_index)
            local_Mi = self.get_number_of_allowed_species_by_site(
                self._get_primitive_structure(), orbit.representative_sites)
            mc_vectors = orbit.get_mc_vectors(local_Mi)
            mc_permutations = self.get_multi_component_vector_permutations(
                mc_vectors, orbit_index)
            mc_index = mc_vectors.index(mc_vector)
            mc_permutations_multiplicity = len(mc_permutations[mc_index])
            cluster = self.get_orbit(orbit_index).get_representative_cluster()
            multiplicity = len(self.get_orbit(
                orbit_index).get_equivalent_sites())
            record = OrderedDict([('index', index),
                                  ('order', cluster.order),
                                  ('radius', cluster.radius),
                                  ('multiplicity', multiplicity *
                                   mc_permutations_multiplicity),
                                  ('orbit_index', orbit_index)])
            record['multi_component_vector'] = mc_vector
            data.append(record)
            index += 1
        return data

    def get_number_of_orbits_by_order(self) -> OrderedDict:
        """
        Return the number of orbits by order.

        Returns
        -------
        dictionary (ordered)
            the key represents the order, the value represents the number of
            orbits
        """
        count_orbits = {}  # type: Dict[int, int]
        for orbit in self.orbit_data:
            k = orbit['order']
            count_orbits[k] = count_orbits.get(k, 0) + 1
        return OrderedDict(sorted(count_orbits.items()))

    def get_cluster_vector(self, atoms: Atoms) -> np.ndarray:
        """
        Returns the cluster vector for a structure.

        Parameters
        ----------
        atoms
            atomic configuration

        Returns
        -------
        the cluster vector
        """
        assert isinstance(atoms, Atoms), \
            'input configuration must be an ASE Atoms object'
        if not atoms.pbc.all():
            add_vacuum_in_non_pbc(atoms)
        return _ClusterSpace.get_cluster_vector(self,
                                                Structure.from_atoms(atoms))

    def _prune_orbit_list(self, indices: List[int]) -> None:
        """
        Prunes the internal orbit list

        Parameters
        ----------
        indices
            indices to all orbits to be removed
        """
        size_before = len(self._orbit_list)

        self._prune_orbit_list_cpp(indices)

        for index in sorted(indices, reverse=True):
            self._orbit_list.remove_orbit(index)
        self._precompute_multi_component_vectors()
        size_after = len(self._orbit_list)
        assert size_before - len(indices) == size_after

    @property
    def primitive_structure(self) -> Atoms:
        """
        Primitive structure on which the cluster space is based
        """
        return self._get_primitive_structure().to_atoms()

    @property
    def chemical_symbols(self) -> List[List[str]]:
        """
        Chemical species considered
        """
        return self._primitive_chemical_symbols.copy()

    @property
    def cutoffs(self) -> List[float]:
        """
        Cutoffs for the different n-body clusters. Each cutoff radii
        (in Angstroms) defines the largest inter-atomic distance in each
        cluster
        """
        return self._cutoffs

    @property
    def orbit_list(self):
        """Orbit list that defines the cluster in the cluster space"""
        return self._orbit_list

    def write(self, filename: str):
        """
        Saves cluster space to a file.

        Parameters
        ---------
        filename
            name of file to which to write
        """

        parameters = {'atoms': self._atoms.copy(),
                      'cutoffs': self._cutoffs,
                      'chemical_symbols': self._input_chemical_symbols}
        with open(filename, 'wb') as handle:
            pickle.dump(parameters, handle, protocol=pickle.HIGHEST_PROTOCOL)

    @staticmethod
    def read(filename: str):
        """
        Reads cluster space from filename.

        Parameters
        ---------
        filename
            name of file from which to read cluster space
        """
        if isinstance(filename, str):
            with open(filename, 'rb') as handle:
                parameters = pickle.load(handle)
        else:
            parameters = pickle.load(filename)

        return ClusterSpace(parameters['atoms'],
                            parameters['cutoffs'],
                            parameters['chemical_symbols'])


def get_singlet_info(atoms: Atoms, return_cluster_space: bool = False):
    """
    Retrieves information concerning the singlets in the input structure.

    Parameters
    ----------
    atoms
        atomic configuration
    return_cluster_space
        if True return the cluster space created during the process

    Returns
    -------
    list of dicts
        each dictionary in the list represents one orbit
    ClusterSpace object (optional)
        cluster space created during the process
    """
    assert isinstance(atoms, Atoms), \
        'input configuration must be an ASE Atoms object'

    # create dummy species and cutoffs
    chemical_symbols = ['H', 'He']
    cutoffs = [0.0]

    cs = ClusterSpace(atoms, cutoffs, chemical_symbols)

    singlet_data = []

    for i in range(1, len(cs)):
        cluster_space_info = cs.get_cluster_space_info(i)
        orbit_index = cluster_space_info[0]
        cluster = cs.get_orbit(orbit_index).get_representative_cluster()
        multiplicity = len(cs.get_orbit(orbit_index).get_equivalent_sites())
        assert len(cluster) == 1, \
            'Cluster space contains higher-order terms (beyond singlets)'

        singlet = {}
        singlet['orbit_index'] = orbit_index
        singlet['sites'] = cs.get_orbit(orbit_index).get_equivalent_sites()
        singlet['multiplicity'] = multiplicity
        singlet['representative_site'] = cs.get_orbit(
            orbit_index).get_representative_sites()
        singlet_data.append(singlet)

    if return_cluster_space:
        return singlet_data, cs
    else:
        return singlet_data


def get_singlet_configuration(atoms: Atoms,
                              to_primitive: bool = False) -> Atoms:
    """
    Returns the atomic configuration decorated with a different species for
    each Wyckoff site. This is useful for visualization and analysis.

    Parameters
    ----------
    atoms
        atomic configuration
    to_primitive
        if True the input structure will be reduced to its primitive unit cell
        before processing

    Returns
    -------
    ASE Atoms object
        structure with singlets highlighted by different chemical species
    """
    from ase.data import chemical_symbols
    assert isinstance(atoms, Atoms), \
        'input configuration must be an ASE Atoms object'
    cluster_data, cluster_space = get_singlet_info(atoms,
                                                   return_cluster_space=True)

    if to_primitive:
        singlet_configuration = cluster_space.primitive_structure
        for singlet in cluster_data:
            for site in singlet['sites']:
                symbol = chemical_symbols[singlet['orbit_index'] + 1]
                atom_index = site[0].index
                singlet_configuration[atom_index].symbol = symbol
    else:
        singlet_configuration = atoms.copy()
        singlet_configuration = add_vacuum_in_non_pbc(singlet_configuration)
        orbit_list_supercell\
            = cluster_space._orbit_list.get_supercell_orbit_list(
                singlet_configuration)
        for singlet in cluster_data:
            for site in singlet['sites']:
                symbol = chemical_symbols[singlet['orbit_index'] + 1]
                sites = orbit_list_supercell.get_orbit(
                    singlet['orbit_index']).get_equivalent_sites()
                for lattice_site in sites:
                    k = lattice_site[0].index
                    singlet_configuration[k].symbol = symbol

    return singlet_configuration


def view_singlets(atoms: Atoms, to_primitive: bool = False):
    """
    Visualize singlets in a structure using the ASE graphical user interface.

    Parameters
    ----------
    atoms
        atomic configuration
    to_primitive
        if True the input structure will be reduced to its primitive unit cell
        before processing
    """
    from ase.visualize import view
    assert isinstance(atoms, Atoms), \
        'input configuration must be an ASE Atoms object'
    singlet_configuration = get_singlet_configuration(
        atoms, to_primitive=to_primitive)
    view(singlet_configuration)



from .convex_hull import ConvexHull
from .structure_enumeration import (enumerate_structures,
                                    get_symmetry_operations)
from .geometry import get_primitive_structure
from .structure_mapping import map_structure_to_reference

__all__ = ['ConvexHull',
           'enumerate_structures',
           'get_symmetry_operations',
           'get_primitive_structure',
           'map_structure_to_reference']



import copy
import random
from numpy import array
from typing import Dict, List, Tuple, Union
from ase import Atoms


class SwapNotPossibleError(Exception):
    pass


class ConfigurationManager(object):
    """
    The ConfigurationManager owns and handles information pertaining to a
    configuration being sampled in a Monte Carlo simulation.

    Parameters
    ----------
    atoms : ASE Atoms
        configuration to be handled
    strict_constraints : list of list of int
        strictest form of the allowed occupations
    sites_by_sublattice : list of list of int
        sites (inner list) that belong to each sublattice (outer list)
    occupation_constraints : list of list of int
        optional occupation constraint to enfore a more stricter species
        occupation than what is allowed from the Calculator

    Todo
    ----
    * occupation constraint not implemented
    * add check that all sites in the different sublattices all have the same
      occupation constraint.
    * revise docstrings
    * clarify "occupation_constraints" vs "strict_constraints";
      the OccupationConstraints class should help here
    """

    def __init__(self, atoms: Atoms,
                 strict_constraints: Union[List[List[int]], List[int]],
                 sites_by_sublattice: Union[List[List[int]], List[int]],
                 occupation_constraints: List[List[int]] = None) -> None:

        self._atoms = atoms
        self._occupations = atoms.numbers
        self._sites_by_sublattice = sites_by_sublattice

        if occupation_constraints is not None:
            self._check_occupation_constraint(
                strict_constraints, occupation_constraints)
        else:
            occupation_constraints = strict_constraints
        self._occupation_constraints = occupation_constraints
        self._allowed_species = self._set_up_allowed_species()
        self._sites_by_species = self._get_sites_by_species()

    def _set_up_allowed_species(self) -> List[int]:
        """ Returns a list of allowed species. """
        allowed_species = set()
        for occ in self._occupation_constraints:
            for species in occ:
                allowed_species.add(species)
        return list(allowed_species)

    def _get_sites_by_species(self) -> List[Dict[int, List[int]]]:
        """Returns the sites that are occupied for each species.  Each
        dictionary represents one sublattice where the key is the
        species (by atomic number) and the value is the list of sites
        occupied by said species in the respective sublattice.
        """
        sites_by_species = []
        for sublattice in self._sites_by_sublattice:
            species_dict = {key: [] for key in self._allowed_species}
            for site in sublattice:
                species_dict[self._occupations[site]].append(site)
            sites_by_species.append(species_dict)
        return sites_by_species

    def _check_occupation_constraint(self,
                                     strict_constraints: Union[List[List[int]],
                                                               List[int]],
                                     occupation_constraints: List[List[int]]):
        """Checks that the user defined occupation constraints are stricter or
        as strict as the strict constraints.

        Parameters
        ----------
        strict_constraints
            additional (stricter) constraints specified for this
            configuration manager instance
        occupation_constraints
            "default" constraints

        Todo
        ----
        This method should be revised and rewritten once the
        OccupationConstraints class is available. At least some of
        this functionality should probably be moved into said class.
        """

        if not len(strict_constraints) == len(occupation_constraints):
            raise ValueError(
                'strict_occupations and occupation_constraints'
                ' must be equal length')

        for strict_occ, occ in zip(strict_constraints, occupation_constraints):
            if not set(occ).issubset(strict_occ):
                raise ValueError(
                    'User defined occupation_constraints must be '
                    'stricter or as strict as strict_occupations constraints.')

    @property
    def occupations(self) -> List[int]:
        """ occupation vector of the configuration (copy) """
        return self._occupations.copy()

    @property
    def occupation_constraints(self) ->List[List[int]]:
        """occupation constraints associated with configuration manager
        (copy); the outer list runs over sites with each inner list
        representing the species allowed on this site.
        """
        return copy.deepcopy(self._occupation_constraints)

    @property
    def sublattices(self) -> List[List[int]]:
        """ sites belonging to each sublattice of the configuration (copy) """
        return copy.deepcopy(self._sites_by_sublattice)

    @property
    def atoms(self) -> Atoms:
        """ atomic structure associated with configuration (copy) """
        atoms = self._atoms.copy()
        atoms.set_atomic_numbers(self.occupations)
        return atoms

    def get_swapped_state(self, sublattice: int) -> Tuple[List[int],
                                                          List[int]]:
        """Returns two random sites (first element of tuple) and their
        occupation after a swap (second element of tuple).  The new
        configuration will obey the occupation constraints associated
        with the configuration mananger.

        Parameters
        ----------
        sublattice
            sublattice from which to pick sites

        Todo
        ----
        * profile this method as it is called frequently
        * look for speed up opportunities
        * The current implementation assumes all sites in this sublattice to
          have the same allowed occupations.
        """
        # pick the first site
        try:
            site1 = random.choice(self._sites_by_sublattice[sublattice])
        except IndexError:
            raise SwapNotPossibleError('Sublattice {} is empty.'
                                       .format(sublattice))

        # pick the second site
        possible_swap_species = \
            set(self._occupation_constraints[site1]) - \
            set([self._occupations[site1]])
        possible_swap_sites = []
        for Z in possible_swap_species:
            possible_swap_sites.extend(self._sites_by_species[sublattice][Z])

        possible_swap_sites = array(possible_swap_sites)

        try:
            site2 = random.choice(possible_swap_sites)
        except IndexError:
            raise SwapNotPossibleError

        return ([site1, site2],
                [self._occupations[site2], self._occupations[site1]])

    def get_flip_state(self, sublattice: int) -> Tuple[int, int]:
        """
        Returns a site index and a new species for the site.

        Parameters
        ----------
        sublattice
            index of sublattice from which to pick a site
        """

        site = random.choice(self._sites_by_sublattice[sublattice])
        species = random.choice(list(
            set(self._occupation_constraints[site]) -
            set([self._occupations[site]])))
        return site, species

    def update_occupations(self, sites: List[int], species: List[int]):
        """
        Updates the occupation vector of the configuration being sampled.
        This will change the state in both the configuration in the calculator
        and the configuration manager.

        Parameters
        ----------
        sites
            indices of sites of the configuration to change
        species
            new occupations by atomic number
        """

        # Update _sites_by_sublattice
        for site, new_Z in zip(sites, species):
            old_Z = self._occupations[site]
            for isub, sublattice_sites in enumerate(self._sites_by_sublattice):
                if site in sublattice_sites:
                    break
            else:
                raise ValueError(
                    'Site {} is not present in any sublattice.'.format(site))

            # Remove site from list of sites for old species
            self._sites_by_species[isub][old_Z].remove(site)
            # Add site to list of sites for new species
            try:
                self._sites_by_species[isub][new_Z].append(site)
            except KeyError:
                raise ValueError('Invalid new species {} on site {}'
                                 .format(new_Z, site))

        # Update occupation vector itself
        self._occupations[sites] = species
